# Copyright 2022 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
from abc import abstractmethod
from typing import Dict, Iterable, Optional, Tuple, TypeVar, Union, cast

import numpy as np
import pandas as pd
from typing_extensions import Protocol

from lale.expressions import astype, count, it, sum
from lale.helpers import spark_installed
from lale.operators import TrainedOperator

from .aggregate import Aggregate
from .map import Map
from .monoid import Monoid, MonoidFactory

MetricMonoid = Monoid

_M = TypeVar("_M", bound=MetricMonoid)

if spark_installed:
    from pyspark.sql.dataframe import DataFrame as SparkDataFrame

    _Batch_Xy = Union[
        Tuple[pd.DataFrame, pd.Series],
        Tuple[SparkDataFrame, SparkDataFrame],
    ]

    _Batch_yyX = Tuple[
        Union[pd.Series, np.ndarray, SparkDataFrame],
        Union[pd.Series, np.ndarray, SparkDataFrame],
        Union[pd.DataFrame, SparkDataFrame],
    ]

else:
    _Batch_Xy = Tuple[pd.DataFrame, pd.Series]  # type: ignore

    _Batch_yyX = Tuple[  # type: ignore
        Union[pd.Series, np.ndarray], Union[pd.Series, np.ndarray], pd.DataFrame
    ]


class MetricMonoidFactory(MonoidFactory[_Batch_yyX, float, _M], Protocol):
    @abstractmethod
    def to_monoid(self, v: _Batch_yyX) -> _M:
        pass

    @abstractmethod
    def score_data(
        self, y_true: pd.Series, y_pred: pd.Series, X: Optional[pd.DataFrame] = None
    ) -> float:
        pass  # keeping this abstract to allow inheriting non-batched version

    @abstractmethod
    def score_estimator(
        self, estimator: TrainedOperator, X: pd.DataFrame, y: pd.Series
    ) -> float:
        pass  # keeping this abstract to allow inheriting non-batched version

    def __call__(
        self, estimator: TrainedOperator, X: pd.DataFrame, y: pd.Series
    ) -> float:
        return self.score_estimator(estimator, X, y)

    def score_data_batched(self, batches: Iterable[_Batch_yyX]) -> float:
        lifted_batches = (self.to_monoid(b) for b in batches)
        combined = functools.reduce(lambda a, b: a.combine(b), lifted_batches)
        return self.from_monoid(combined)

    def score_estimator_batched(
        self, estimator: TrainedOperator, batches: Iterable[_Batch_Xy]
    ) -> float:
        predicted_batches = ((y, estimator.predict(X), X) for X, y in batches)
        return self.score_data_batched(predicted_batches)


class _MetricMonoidMixin(MetricMonoidFactory[_M], Protocol):
    def score_data(
        self, y_true: pd.Series, y_pred: pd.Series, X: Optional[pd.DataFrame] = None
    ) -> float:
        return self.from_monoid(self.to_monoid((y_true, y_pred, X)))

    def score_estimator(
        self, estimator: TrainedOperator, X: pd.DataFrame, y: pd.Series
    ) -> float:
        return self.score_data(y_true=y, y_pred=estimator.predict(X))


def _make_dataframe_yy(batch):
    def make_series_y(y):
        if isinstance(y, np.ndarray):
            series = pd.Series(y)
        elif spark_installed and isinstance(y, SparkDataFrame):
            series = cast(pd.DataFrame, y.toPandas()).squeeze()
        else:
            series = y
        assert isinstance(series, pd.Series), type(series)
        return series.reset_index(drop=True)

    y_true, y_pred, _ = batch
    result = pd.DataFrame(
        {"y_true": make_series_y(y_true), "y_pred": make_series_y(y_pred)},
    )
    return result


class _AccuracyData(MetricMonoid):
    def __init__(self, match: int, total: int):
        self.match = match
        self.total = total

    def combine(self, other: "_AccuracyData") -> "_AccuracyData":
        return _AccuracyData(self.match + other.match, self.total + other.total)


class _Accuracy(_MetricMonoidMixin[_AccuracyData]):
    def __init__(self):
        self._pipeline = Map(
            columns={"match": astype("int", it.y_true == it.y_pred)}
        ) >> Aggregate(columns={"match": sum(it.match), "total": count(it.match)})

    def to_monoid(self, batch: _Batch_yyX) -> _AccuracyData:
        input_df = _make_dataframe_yy(batch)
        agg_df = self._pipeline.transform(input_df)
        return _AccuracyData(match=agg_df.at[0, "match"], total=agg_df.at[0, "total"])

    def from_monoid(self, v: _AccuracyData) -> float:
        return float(v.match / np.float64(v.total))


def accuracy_score(y_true: pd.Series, y_pred: pd.Series) -> float:
    return get_scorer("accuracy").score_data(y_true, y_pred)


class _R2Data(MetricMonoid):
    def __init__(self, n: int, sum: float, sum_sq: float, res_sum_sq: float):
        self.n = n
        self.sum = sum
        self.sum_sq = sum_sq
        self.res_sum_sq = res_sum_sq

    def combine(self, other: "_R2Data") -> "_R2Data":
        return _R2Data(
            n=self.n + other.n,
            sum=self.sum + other.sum,
            sum_sq=self.sum_sq + other.sum_sq,
            res_sum_sq=self.res_sum_sq + other.res_sum_sq,
        )


class _R2(_MetricMonoidMixin[_R2Data]):
    # https://en.wikipedia.org/wiki/Coefficient_of_determination

    def __init__(self):
        self._pipeline = Map(
            columns={
                "y": it.y_true,  # observed values
                "f": it.y_pred,  # predicted values
                "y2": it.y_true * it.y_true,  # squares
                "e2": (it.y_true - it.y_pred) * (it.y_true - it.y_pred),
            }
        ) >> Aggregate(
            columns={
                "n": count(it.y),
                "sum": sum(it.y),
                "sum_sq": sum(it.y2),
                "res_sum_sq": sum(it.e2),  # residual sum of squares
            }
        )

    def to_monoid(self, batch: _Batch_yyX) -> _R2Data:
        input_df = _make_dataframe_yy(batch)
        agg_df = self._pipeline.transform(input_df)
        return _R2Data(
            n=agg_df.at[0, "n"],
            sum=agg_df.at[0, "sum"],
            sum_sq=agg_df.at[0, "sum_sq"],
            res_sum_sq=agg_df.at[0, "res_sum_sq"],
        )

    def from_monoid(self, v: _R2Data) -> float:
        ss_tot = v.sum_sq - (v.sum * v.sum / np.float64(v.n))
        return 1 - float(v.res_sum_sq / ss_tot)


def r2_score(y_true: pd.Series, y_pred: pd.Series) -> float:
    return get_scorer("r2").score_data(y_true, y_pred)


_scorer_cache: Dict[str, Optional[MetricMonoidFactory]] = {"accuracy": None, "r2": None}


def get_scorer(scoring: str) -> MetricMonoidFactory:
    assert scoring in _scorer_cache, scoring
    if _scorer_cache[scoring] is None:
        if scoring == "accuracy":
            _scorer_cache[scoring] = _Accuracy()
        elif scoring == "r2":
            _scorer_cache[scoring] = _R2()
    result = _scorer_cache[scoring]
    assert result is not None
    return result



# Copyright 2022 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
try:
    import snapml  # type: ignore

    snapml_installed = True
except ImportError:
    snapml_installed = False

import lale.datasets.data_schemas
import lale.docstrings
import lale.operators


class _BatchedTreeEnsembleClassifierImpl:
    def __init__(
        self,
        base_ensemble=None,
        max_sub_ensembles=10,
        inner_lr_scaling=0.5,
        outer_lr_scaling=0.5,
    ):
        assert (
            snapml_installed
        ), """Your Python environment does not have snapml installed. Install using: pip install snapml"""
        if base_ensemble is None:
            from snapml import SnapBoostingMachineClassifier

            base_ensemble = SnapBoostingMachineClassifier()
        self._hyperparams = {
            "base_ensemble": base_ensemble,
            "max_sub_ensembles": max_sub_ensembles,
            "inner_lr_scaling": inner_lr_scaling,
            "outer_lr_scaling": outer_lr_scaling,
        }
        self._wrapped_model = snapml.BatchedTreeEnsembleClassifier(**self._hyperparams)

    def fit(self, X, y, **fit_params):
        X = lale.datasets.data_schemas.strip_schema(X)
        y = lale.datasets.data_schemas.strip_schema(y)
        self._wrapped_model.fit(X, y, **fit_params)
        return self

    def predict(self, X, **predict_params):
        X = lale.datasets.data_schemas.strip_schema(X)
        return self._wrapped_model.predict(X, **predict_params)

    def predict_proba(self, X, **predict_proba_params):
        X = lale.datasets.data_schemas.strip_schema(X)
        return self._wrapped_model.predict_proba(X, **predict_proba_params)

    def partial_fit(self, X, y, **fit_params):
        X = lale.datasets.data_schemas.strip_schema(X)
        y = lale.datasets.data_schemas.strip_schema(y)
        self._wrapped_model.partial_fit(X, y, **fit_params)
        return self


_hyperparams_schema = {
    "description": "Hyperparameter schema.",
    "allOf": [
        {
            "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.",
            "type": "object",
            "additionalProperties": True,
            "relevantToOptimizer": [],
            "properties": {},
        }
    ],
}

_input_fit_schema = {
    "description": "Fit the base ensemble without batching.",
    "type": "object",
    "required": ["X", "y"],
    "properties": {
        "X": {
            "type": "array",
            "description": "The outer array is over samples aka rows.",
            "items": {
                "type": "array",
                "description": "The inner array is over features aka columns.",
                "items": {"type": "number"},
            },
        },
        "y": {
            "description": "The classes.",
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {"type": "array", "items": {"type": "string"}},
                {"type": "array", "items": {"type": "boolean"}},
            ],
        },
        "sample_weight": {
            "anyOf": [
                {"type": "array", "items": {"type": "number"}},
                {"enum": [None], "description": "Samples are equally weighted."},
            ],
            "description": "Sample weights.",
            "default": None,
        },
    },
}

_input_predict_schema = {
    "type": "object",
    "required": ["X"],
    "properties": {
        "X": {
            "type": "array",
            "description": "The outer array is over samples aka rows.",
            "items": {
                "type": "array",
                "description": "The inner array is over features aka columns.",
                "items": {"type": "number"},
            },
        },
    },
}

_output_predict_schema = {
    "description": "The predicted classes.",
    "anyOf": [
        {"type": "array", "items": {"type": "number"}},
        {"type": "array", "items": {"type": "string"}},
        {"type": "array", "items": {"type": "boolean"}},
    ],
}

_input_predict_proba_schema = {
    "type": "object",
    "properties": {
        "X": {
            "type": "array",
            "description": "The outer array is over samples aka rows.",
            "items": {
                "type": "array",
                "description": "The inner array is over features aka columns.",
                "items": {"type": "number"},
            },
        },
    },
}

_output_predict_proba_schema = {
    "type": "array",
    "description": "The outer array is over samples aka rows.",
    "items": {
        "type": "array",
        "description": "The inner array contains probabilities corresponding to each class.",
        "items": {"type": "number"},
    },
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """`Batched Tree Ensemble Classifier`_ from `Snap ML`_.

.. _`Batched Tree Ensemble Classifier`: https://snapml.readthedocs.io/en/latest/batched_tree_ensembles.html
.. _`Snap ML`: https://www.zurich.ibm.com/snapml/
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.snapml.batched_tree_ensemble_classifier.html",
    "import_from": "snapml",
    "type": "object",
    "tags": {"pre": [], "op": ["estimator", "classifier"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_fit": _input_fit_schema,
        "input_predict": _input_predict_schema,
        "output_predict": _output_predict_schema,
        "input_predict_proba": _input_predict_proba_schema,
        "output_predict_proba": _output_predict_proba_schema,
    },
}


BatchedTreeEnsembleClassifier = lale.operators.make_operator(
    _BatchedTreeEnsembleClassifierImpl, _combined_schemas
)

lale.docstrings.set_docstrings(BatchedTreeEnsembleClassifier)



# Copyright 2022 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable, Tuple, Union, cast

import pandas as pd
import sklearn.model_selection
import sklearn.tree

import lale.helpers
from lale.datasets import pandas2spark

from .split_xy import SplitXy

_PandasBatch = Tuple[pd.DataFrame, pd.Series]

if lale.helpers.spark_installed:
    from pyspark.sql.dataframe import DataFrame as SparkDataFrame

    _PandasOrSparkBatch = Union[
        Tuple[pd.DataFrame, pd.Series],
        Tuple[SparkDataFrame, SparkDataFrame],
    ]
else:
    _PandasOrSparkBatch = _PandasBatch  # type: ignore

try:
    import arff

    from lale.datasets.openml import openml_datasets

    liac_arff_installed = True
except ModuleNotFoundError:
    liac_arff_installed = False


def arff_data_loader(
    file_name: str, label_name: str, rows_per_batch: int
) -> Iterable[_PandasBatch]:
    """Incrementally load an ARFF file and yield it one (X, y) batch at a time."""
    assert liac_arff_installed
    split_x_y = SplitXy(label_name=label_name)

    def make_batch():
        start = n_batches * rows_per_batch
        stop = start + len(row_list)
        df = pd.DataFrame(row_list, range(start, stop), column_names)
        X, y = split_x_y.transform_X_y(df, None)
        return X, y

    with open(file_name) as f:
        arff_dict = arff.load(f, return_type=arff.DENSE_GEN)
        column_names = [name for name, _ in arff_dict["attributes"]]
        row_list = []
        n_batches = 0
        for row in arff_dict["data"]:
            row_list.append(row)
            if len(row_list) >= rows_per_batch:
                yield make_batch()
                row_list = []
                n_batches += 1
    if len(row_list) > 0:  # last chunk
        yield make_batch()


def csv_data_loader(
    file_name: str, label_name: str, rows_per_batch: int
) -> Iterable[_PandasBatch]:
    """Incrementally load an CSV file and yield it one (X, y) batch at a time."""
    split_x_y = SplitXy(label_name=label_name)
    with pd.read_csv(file_name, chunksize=rows_per_batch) as reader:
        for df in reader:
            X, y = split_x_y.transform_X_y(df, None)
            yield X, y


def mockup_data_loader(
    X: pd.DataFrame, y: pd.Series, n_batches: int, astype: str
) -> Iterable[_PandasOrSparkBatch]:
    """Split (X, y) into batches to emulate loading them incrementally.

    Only intended for testing purposes, because if X and y are already
    materialized in-memory, there is little reason to batch them.
    """
    pandas_gen: Iterable[_PandasBatch]
    if n_batches == 1:
        pandas_gen = [(X, y)]
    else:
        cv = sklearn.model_selection.KFold(n_batches)
        estimator = sklearn.tree.DecisionTreeClassifier()
        pandas_gen = (
            lale.helpers.split_with_schemas(estimator, X, y, test, train)
            for train, test in cv.split(X, y)
        )
    if astype == "pandas":
        return pandas_gen
    elif astype == "spark":
        return ((pandas2spark(X), pandas2spark(y)) for X, y in pandas_gen)
    raise ValueError(f"expected astype in ['pandas', 'spark'], got {astype}")


def openml_data_loader(dataset_name: str, batch_size: int) -> Iterable[_PandasBatch]:
    """Download the OpenML dataset, incrementally load it, and yield it one (X,y) batch at a time."""
    assert liac_arff_installed
    metadata = openml_datasets.experiments_dict[dataset_name]
    label_name = cast(str, metadata["target"])
    file_name = openml_datasets.download_if_missing(dataset_name)
    return arff_data_loader(file_name, label_name, batch_size)



# # Copyright 2022 IBM Corporation
# #
# # Licensed under the Apache License, Version 2.0 (the "License");
# # you may not use this file except in compliance with the License.
# # You may obtain a copy of the License at
# #
# # http://www.apache.org/licenses/LICENSE-2.0
# #
# # Unless required by applicable law or agreed to in writing, software
# # distributed under the License is distributed on an "AS IS" BASIS,
# # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# # See the License for the specific language governing permissions and
# # limitations under the License.

# import pandas as pd
# from autoai_libs.transformers.date_time.date_time_transformer import (
#     DateTransformer as model_to_be_wrapped,
# )

# import lale.docstrings
# import lale.operators


# class _DateTransformerImpl:
#     def __init__(
#         self,
#         options=None,
#         delete_source_columns=False,
#         column_headers_list=None,
#         missing_values_reference_list=None,
#         activate_flag=True,
#         float32_processing_flag=True,
#     ):
#         self._hyperparams = {
#             "options": options,
#             "delete_source_columns": delete_source_columns,
#             "column_headers_list": column_headers_list,
#             "missing_values_reference_list": missing_values_reference_list,
#             "activate_flag": activate_flag,
#             "float32_processing_flag": float32_processing_flag,
#         }
#         self._wrapped_model = model_to_be_wrapped(**self._hyperparams)

#     def fit(self, X, y=None):
#         if isinstance(X, pd.DataFrame):
#             X = X.to_numpy()
#         self._wrapped_model.fit(X, y)
#         return self

#     def transform(self, X):
#         if isinstance(X, pd.DataFrame):
#             X = X.to_numpy()
#         return self._wrapped_model.transform(X)


# _hyperparams_schema = {
#     "allOf": [
#         {
#             "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters.",
#             "type": "object",
#             "additionalProperties": False,
#             "required": [
#                 "options",
#                 "delete_source_columns",
#                 "column_headers_list",
#                 "missing_values_reference_list",
#                 "activate_flag",
#                 "float32_processing_flag",
#             ],
#             "relevantToOptimizer": [],
#             "properties": {
#                 "options": {
#                     "description": """List containing the types of new feature columns to add for each detected datetime column.
# Default is None, in this case all the above options are applied""",
#                     "anyOf": [
#                         {
#                             "type": "array",
#                             "items": {
#                                 "enum": [
#                                     "all",
#                                     "Datetime",
#                                     "DateToFloatTimestamp",
#                                     "DateToTimestamp",
#                                     "Timestamp",
#                                     "FloatTimestamp",
#                                     "DayOfWeek",
#                                     "DayOfMonth",
#                                     "Hour",
#                                     "DayOfYear",
#                                     "Week",
#                                     "Month",
#                                     "Year",
#                                     "Second",
#                                     "Minute",
#                                 ]
#                             },
#                         },
#                         {"enum": [None]},
#                     ],
#                     "default": None,
#                 },
#                 "delete_source_columns": {
#                     "description": "Flag determining whether the original date columns will be deleted or not.",
#                     "type": "boolean",
#                     "default": True,
#                 },
#                 "column_headers_list": {
#                     "description": "List containing the column names of the input array",
#                     "anyOf": [
#                         {
#                             "type": "array",
#                             "items": {
#                                 "anyOf": [{"type": "string"}, {"type": "integer"}]
#                             },
#                         },
#                         {"enum": [None]},
#                     ],
#                     "default": None,
#                 },
#                 "missing_values_reference_list": {
#                     "description": "List containing missing values of the input array",
#                     "anyOf": [
#                         {"type": "array", "items": {"laleType": "Any"}},
#                         {"enum": [None]},
#                     ],
#                     "default": None,
#                 },
#                 "activate_flag": {
#                     "description": "Determines whether transformer is active or not.",
#                     "type": "boolean",
#                     "default": True,
#                 },
#                 "float32_processing_flag": {
#                     "description": "Flag that determines whether timestamps will be float32-compatible.",
#                     "type": "boolean",
#                     "default": True,
#                 },
#             },
#         }
#     ]
# }

# _input_fit_schema = {
#     "type": "object",
#     "required": ["X"],
#     "additionalProperties": False,
#     "properties": {
#         "X": {  # Handles 1-D arrays as well
#             "anyOf": [
#                 {"type": "array", "items": {"laleType": "Any"}},
#                 {
#                     "type": "array",
#                     "items": {"type": "array", "items": {"laleType": "Any"}},
#                 },
#             ]
#         },
#         "y": {"laleType": "Any"},
#     },
# }

# _input_transform_schema = {
#     "type": "object",
#     "required": ["X"],
#     "additionalProperties": False,
#     "properties": {
#         "X": {  # Handles 1-D arrays as well
#             "anyOf": [
#                 {"type": "array", "items": {"laleType": "Any"}},
#                 {
#                     "type": "array",
#                     "items": {"type": "array", "items": {"laleType": "Any"}},
#                 },
#             ]
#         }
#     },
# }

# _output_transform_schema = {
#     "description": "Features; the outer array is over samples.",
#     "anyOf": [
#         {"type": "array", "items": {"laleType": "Any"}},
#         {"type": "array", "items": {"type": "array", "items": {"laleType": "Any"}}},
#     ],
# }

# _combined_schemas = {
#     "$schema": "http://json-schema.org/draft-04/schema#",
#     "description": """Operator from `autoai_libs`_. Detects date columns on an input array and adds new feature columns for each detected date column.

# .. _`autoai_libs`: https://pypi.org/project/autoai-libs""",
#     "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.autoai_libs.date_transformer.html",
#     "import_from": "autoai_libs.transformers.date_time",
#     "type": "object",
#     "tags": {"pre": [], "op": ["transformer"], "post": []},
#     "properties": {
#         "hyperparams": _hyperparams_schema,
#         "input_fit": _input_fit_schema,
#         "input_transform": _input_transform_schema,
#         "output_transform": _output_transform_schema,
#     },
# }


# DateTransformer = lale.operators.make_operator(_DateTransformerImpl, _combined_schemas)

# lale.docstrings.set_docstrings(DateTransformer)



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Tuple, TypeVar

import numpy as np
from scipy import special
from typing_extensions import Protocol

from lale.expressions import count as agg_count
from lale.expressions import it
from lale.expressions import sum as agg_sum
from lale.helpers import _ensure_pandas, _is_pandas_series
from lale.lib.dataframe import get_columns
from lale.lib.rasl import Aggregate, ConcatFeatures, GroupBy, Map

from .monoid import Monoid, MonoidFactory

ScoreMonoid = Monoid

_InputType = Tuple[Any, Any]  # TODO: be more precise?
_OutputType = Tuple[float, float]
_M = TypeVar("_M", bound=ScoreMonoid)


class ScoreMonoidFactory(MonoidFactory[_InputType, _OutputType, _M], Protocol):
    def score(self, X, y) -> Tuple[float, float]:
        return self.from_monoid(self.to_monoid((X, y)))


class FOnewayData(Monoid):
    def __init__(
        self,
        *,
        classes,
        n_samples_per_class,
        n_samples,
        ss_alldata,
        sums_samples,
        sums_alldata,
    ):
        """
        Parameters
        ----------
        classes: list
            The list of classes.
        n_samples_per_class: dictionary
            The number of samples in each class.
        n_samples: number
            The total number of samples.
        ss_alldata: array
            The sum of square of each feature.
        sums_samples: dictionary
            The sum of each feaure per class.
        sums_alldata: array
            The sum of each feaure.
        """
        self.classes = classes
        self.n_samples_per_class = n_samples_per_class
        self.n_samples = n_samples
        self.ss_alldata = ss_alldata
        self.sums_samples = sums_samples
        self.sums_alldata = sums_alldata

    def combine(self, other: "FOnewayData"):
        classes_a = self.classes
        n_samples_per_class_a = self.n_samples_per_class
        n_samples_a = self.n_samples
        ss_alldata_a = self.ss_alldata
        sums_samples_a = self.sums_samples
        sums_alldata_a = self.sums_alldata
        classes_b = other.classes
        n_samples_per_class_b = other.n_samples_per_class
        n_samples_b = other.n_samples
        ss_alldata_b = other.ss_alldata
        sums_samples_b = other.sums_samples
        sums_alldata_b = other.sums_alldata
        classes = list(set(classes_a + classes_b))
        n_samples_per_class = {
            k: (n_samples_per_class_a[k] if k in n_samples_per_class_a else 0)
            + (n_samples_per_class_b[k] if k in n_samples_per_class_b else 0)
            for k in classes
        }
        n_samples = n_samples_a + n_samples_b
        ss_alldata = ss_alldata_a + ss_alldata_b
        sums_samples = {
            k: (sums_samples_a[k] if k in sums_samples_a else 0)
            + (sums_samples_b[k] if k in sums_samples_b else 0)
            for k in classes
        }
        sums_alldata = sums_alldata_a + sums_alldata_b
        return FOnewayData(
            classes=classes,
            n_samples_per_class=n_samples_per_class,
            n_samples=n_samples,
            ss_alldata=ss_alldata,
            sums_samples=sums_samples,
            sums_alldata=sums_alldata,
        )


def _gen_name(base, avoid):
    if base not in avoid:
        return base
    cpt = 0
    while f"{base}{cpt}" in avoid:
        cpt += 1
    return f"{base}{cpt}"


# The following function is a rewriting of sklearn.feature_selection.f_oneway
# Compared to the sklearn.feature_selection.f_oneway implementation it
# takes as input the dataset and the target vector.
# Moreover, the function is splitted into two parts: `_f_oneway_lift` and
# `_f_oneway_lower`.
def _f_oneway_lift(X, y) -> FOnewayData:
    """Prepare the data for a 1-way ANOVA.

    Parameters
    ----------
    X: array
        The sample measurements.
    y: array
        The target vector.

    Returns
    -------
    monoid: FOnewayData
        The inermediate data that can be combine for incremental computation.
    """
    if get_columns(y)[0] is None:
        if _is_pandas_series(y):
            y = y.rename(_gen_name("target", get_columns(X)))
    Xy = ConcatFeatures().transform([X, y])
    X_by_y = GroupBy(by=[it[get_columns(y)[0]]]).transform(Xy)

    agg_sum_cols = Aggregate(columns={col: agg_sum(it[col]) for col in get_columns(X)})
    sums_samples = _ensure_pandas(agg_sum_cols.transform(X_by_y))
    n_samples_per_class = Aggregate(
        columns={"n_samples_per_class": agg_count(it[get_columns(X)[0]])}
    ).transform(X_by_y)
    n_samples = _ensure_pandas(
        Aggregate(columns={"sum": agg_sum(it["n_samples_per_class"])}).transform(
            n_samples_per_class
        )
    )["sum"][0]
    sqr_cols = Map(columns={col: it[col] ** 2 for col in get_columns(X)})
    ss_alldata = _ensure_pandas((sqr_cols >> agg_sum_cols).transform(X)).loc[0]
    sums_alldata = _ensure_pandas(agg_sum_cols.transform(X)).loc[0].to_numpy()
    n_samples_per_class = _ensure_pandas(n_samples_per_class).to_dict()[
        "n_samples_per_class"
    ]
    classes = list(n_samples_per_class.keys())
    sums_samples = {k: sums_samples.loc[k].to_numpy() for k in classes}

    return FOnewayData(
        classes=classes,
        n_samples_per_class=n_samples_per_class,
        n_samples=n_samples,
        ss_alldata=ss_alldata,
        sums_samples=sums_samples,
        sums_alldata=sums_alldata,
    )


def _f_oneway_lower(lifted: FOnewayData):
    """Performs a 1-way ANOVA.

    Parameters
    ----------
    lifted : FOnewayData
        The result of `to_monoid`.

    Returns
    -------
    F-value : float
        The computed F-value of the test.
    p-value : float
        The associated p-value from the F-distribution.
    """
    classes = lifted.classes
    n_samples_per_class = lifted.n_samples_per_class
    n_samples = lifted.n_samples
    ss_alldata = lifted.ss_alldata
    sums_samples = lifted.sums_samples
    sums_alldata = lifted.sums_alldata
    n_classes = len(classes)
    square_of_sums_alldata = sums_alldata**2
    square_of_sums_args = {k: s**2 for k, s in sums_samples.items()}
    sstot = ss_alldata - square_of_sums_alldata / float(n_samples)
    ssbn = 0.0
    for k in n_samples_per_class:
        ssbn += square_of_sums_args[k] / n_samples_per_class[k]
    ssbn -= square_of_sums_alldata / float(n_samples)
    sswn = sstot - ssbn
    dfbn = n_classes - 1
    dfwn = n_samples - n_classes
    msb = ssbn / float(dfbn)
    msw = sswn / float(dfwn)
    # constant_features_idx = np.where(msw == 0.0)[0]
    # if (np.nonzero(msb)[0].size != msb.size and constant_features_idx.size):
    #     warnings.warn("Features %s are constant." % constant_features_idx,
    #                   UserWarning)
    f = msb / msw
    # flatten matrix to vector in sparse case
    f = np.asarray(f).ravel()
    prob = special.fdtrc(dfbn, dfwn, f)
    return f, prob


class FClassif(ScoreMonoidFactory[FOnewayData]):
    """Compute the ANOVA F-value for the provided sample."""

    def to_monoid(self, v: Tuple[Any, Any]) -> FOnewayData:
        X, y = v
        return _f_oneway_lift(X, y)

    def from_monoid(self, lifted: FOnewayData):
        return _f_oneway_lower(lifted)



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import reduce
from typing import Any, List, Optional, Tuple

import pandas as pd

import lale.docstrings
import lale.helpers
import lale.operators
from lale.expressions import Expr, hash_mod, it, ite
from lale.helpers import _is_pandas_df, _is_spark_df
from lale.lib.dataframe import count, get_columns
from lale.lib.sklearn import hashing_encoder

from .map import Map
from .monoid import Monoid, MonoidableOperator


# Based on https://github.com/scikit-learn-contrib/category_encoders/blob/master/category_encoders/utils.py
def get_obj_cols(df):
    """
    Returns names of 'object' columns in the DataFrame.
    """
    obj_cols = []
    if _is_pandas_df(df):
        for idx, dt in enumerate(df.dtypes):
            if dt == "object" or is_category(dt):
                obj_cols.append(df.columns.values[idx])
    elif _is_spark_df(df):
        for idx, (col, dt) in enumerate(df.dtypes):
            if dt == "string":
                obj_cols.append(col)
    else:
        assert False

    return obj_cols


# From https://github.com/scikit-learn-contrib/category_encoders/blob/master/category_encoders/utils.py
def is_category(dtype):
    return pd.api.types.is_categorical_dtype(dtype)


class _HashingEncoderMonoid(Monoid):
    def __init__(self, *, n_samples_seen_, feature_names):
        self.n_samples_seen_ = n_samples_seen_
        self.feature_names = feature_names

    def combine(self, other: "_HashingEncoderMonoid"):
        assert list(self.feature_names) == list(other.feature_names)
        n_samples_seen_ = self.n_samples_seen_ + other.n_samples_seen_
        return _HashingEncoderMonoid(
            n_samples_seen_=n_samples_seen_, feature_names=self.feature_names
        )


class _HashingEncoderImpl(MonoidableOperator[_HashingEncoderMonoid]):
    def __init__(
        self,
        *,
        n_components=8,
        cols: Optional[List[str]] = None,
        # drop_invariant=False,
        # return_df=True,
        hash_method="md5",
    ):
        self._hyperparams = {
            "n_components": n_components,
            "cols": cols,
            # "drop_invariant": drop_invariant,
            "hash_method": hash_method,
        }
        self._dim = None

    def transform(self, X):
        if self._transformer is None:
            self._transformer = self._build_transformer(X)
        return self._transformer.transform(X)

    @property
    def n_samples_seen_(self):
        return getattr(self._monoid, "n_samples_seen_", 0)

    @property
    def feature_names(self):
        return getattr(self._monoid, "feature_names", None)

    def from_monoid(self, lifted):
        self._monoid = lifted
        self._transformer = None

    def _build_transformer(self, X):
        cols = self._hyperparams["cols"]
        hash_method = self._hyperparams["hash_method"]
        N = self._hyperparams["n_components"]
        columns_hash = {
            col_name: hash_mod(hash_method, it[col_name], N) for col_name in cols
        }
        columns_cat = {
            f"col_{i}": reduce(
                Expr.__add__,
                [ite(it[col_name] == i, 1, 0) for col_name in cols],
            )
            for i in range(N)
        }
        hash = Map(columns=columns_hash, remainder="passthrough")
        encode = Map(columns=columns_cat, remainder="passthrough")
        return hash >> encode

    def to_monoid(self, v: Tuple[Any, Any]):
        X, y = v
        cols = self._hyperparams["cols"]
        if cols is None:
            cols = get_obj_cols(X)
            self._hyperparams["cols"] = cols

        N = self._hyperparams["n_components"]
        feature_names_cat = [f"col_{i}" for i in range(N)]
        feature_names_num = [col for col in get_columns(X) if col not in cols]
        feature_names = feature_names_cat + feature_names_num  # type: ignore
        n_samples_seen_ = count(X)
        return _HashingEncoderMonoid(
            n_samples_seen_=n_samples_seen_, feature_names=feature_names
        )

    # https://github.com/scikit-learn-contrib/category_encoders/blob/master/category_encoders/hashing.py
    def get_feature_names(self):
        """
        Returns the names of all transformed / added columns.
        Returns
        -------
        feature_names: list
            A list with all feature names transformed or added.
            Note: potentially dropped features are not included!
        """

        if not isinstance(self.feature_names, list):
            raise ValueError(
                "Must fit data first. Affected feature names are not known before."
            )
        else:
            return self.feature_names


_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": """Relational algebra reimplementation of scikit-learn contrib's `HashingEncoder`_ transformer.
Works on both pandas and Spark dataframes by using `Map`_ for `transform`, which in turn use the appropriate backend.

.. _`HashingEncoder`: https://contrib.scikit-learn.org/category_encoders/hashing.html
.. _`Map`: https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.map.html
""",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.hashing_encoder.html",
    "type": "object",
    "tags": {
        "pre": ["categoricals"],
        "op": ["transformer", "interpretable"],
        "post": [],
    },
    "properties": {
        "hyperparams": hashing_encoder._hyperparams_schema,
        "input_fit": hashing_encoder._input_fit_schema,
        "input_transform": hashing_encoder._input_transform_schema,
        "output_transform": hashing_encoder._output_transform_schema,
    },
}

HashingEncoder = lale.operators.make_operator(_HashingEncoderImpl, _combined_schemas)

# HashingEncoder = typing.cast(
#     lale.operators.PlannedIndividualOp,
# )

lale.docstrings.set_docstrings(HashingEncoder)



# Copyright 2019 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import jsonschema
import numpy as np
import pandas as pd

import lale.operators
from lale.lib.rasl.convert import Convert
from lale.operator_wrapper import wrap_imported_operators

try:
    from pyspark import SparkConf, SparkContext
    from pyspark.sql import Row, SQLContext

    from lale.datasets.data_schemas import SparkDataFrameWithIndex

    spark_installed = True
except ImportError:
    spark_installed = False

from test import EnableSchemaValidation

from sklearn.datasets import load_iris
from sklearn.model_selection import train_test_split

from lale.datasets import pandas2spark
from lale.datasets.data_schemas import add_table_name, get_index_name, get_table_name
from lale.datasets.multitable import multitable_train_test_split
from lale.datasets.multitable.fetch_datasets import fetch_go_sales_dataset
from lale.expressions import (
    asc,
    astype,
    collect_set,
    count,
    day_of_month,
    day_of_week,
    day_of_year,
    desc,
    first,
    hour,
    identity,
    isnan,
    isnotnan,
    isnotnull,
    isnull,
    it,
    max,
    mean,
    median,
    min,
    minute,
    mode,
    month,
    replace,
    string_indexer,
    sum,
)
from lale.helpers import (
    _ensure_pandas,
    _is_pandas_df,
    _is_spark_df,
    _is_spark_with_index,
)
from lale.lib.dataframe import get_columns
from lale.lib.lale import ConcatFeatures, Hyperopt, SplitXy
from lale.lib.rasl import (
    Aggregate,
    Alias,
    Filter,
    GroupBy,
    Join,
    Map,
    OrderBy,
    Relational,
    Scan,
)
from lale.lib.sklearn import PCA, KNeighborsClassifier, LogisticRegression


def _set_index_name(df, name):
    return add_table_name(df.rename_axis(index=name), get_table_name(df))


def _set_index(df, name):
    return add_table_name(df.set_index(name), get_table_name(df))


# Testing '==' and '!=' operator with different types of expressions
class TestExpressions(unittest.TestCase):
    def test_expr_1(self):
        with self.assertRaises(TypeError):
            if it.col < 3:
                _ = "If it throws an exception, then the test is successful."

    def test_expr_2(self):
        self.assertFalse(it.col == 5)

    def test_expr_3(self):
        try:
            if it.col == it.col:
                _ = "If it does not throw an exception, then the test is successful."
        except Exception:
            self.fail("Expression 'it.col == it.col' raised an exception unexpectedly!")

    def test_expr_4(self):
        self.assertFalse(it.col == it.col2)

    def test_expr_5(self):
        X = it.col
        self.assertTrue(X == X)

    def test_expr_6(self):
        self.assertFalse(it.col != 5)

    def test_expr_7(self):
        try:
            if it.col != it.col:
                _ = "If it does not throw an exception, then the test is successful."
        except Exception:
            self.fail("Expression 'it.col != it.col' raised an exception unexpectedly!")

    def test_expr_8(self):
        self.assertFalse(it.col != it.col2)

    def test_expr_9(self):
        X = it.col
        self.assertTrue(X != X)


# Testing filter operator
class TestFilter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        info = [
            (1, "NY", 100),
            (2, "NY", 150),
            (3, "TX", 200),
            (4, "TX", 100),
            (5, "CA", 200),
        ]
        t1 = [(2, "Warm"), (3, "Cold"), (4, "Warm"), (5, "Cold")]
        main = [
            (1, "NY", 1, float(1)),
            (2, "TX", 6, np.nan),
            (3, "CA", 2, float(2)),
            (4, "NY", 5, None),
            (5, "CA", 0, float(3)),
        ]

        if spark_installed:
            conf = (
                SparkConf()
                .setMaster("local[2]")
                .set("spark.driver.bindAddress", "127.0.0.1")
            )
            sc = SparkContext.getOrCreate(conf=conf)
            sqlContext = SQLContext(sc)

            rdd = sc.parallelize(main)
            table_main = rdd.map(
                lambda x: Row(TrainId=int(x[0]), col1=x[1], col2=int(x[2]), col6=x[3])
            )
            spark_main = add_table_name(sqlContext.createDataFrame(table_main), "main")

            rdd = sc.parallelize(info)
            table_info = rdd.map(
                lambda x: Row(train_id=int(x[0]), col3=x[1], col4=int(x[2]))
            )
            spark_info = add_table_name(sqlContext.createDataFrame(table_info), "info")

            rdd = sc.parallelize(t1)
            table_t1 = rdd.map(lambda x: Row(tid=int(x[0]), col5=x[1]))
            spark_t1 = add_table_name(sqlContext.createDataFrame(table_t1), "t1")

            trainable = Join(
                pred=[
                    it.main.TrainId == it.info.train_id,
                    it.info.train_id == it.t1.tid,
                ],
                join_type="left",
            )
            spark_transformed_df = trainable.transform(
                [spark_main, spark_info, spark_t1]
            ).sort("TrainId")
            cls.tgt2datasets = {
                "pandas": spark_transformed_df.toPandas(),
                "spark": spark_transformed_df,
                "spark-with-index": SparkDataFrameWithIndex(spark_transformed_df),
            }
        else:
            pandas_main = pd.DataFrame(main, index=["TrainId", "col1", "col2", "col6"])
            pandas_info = pd.DataFrame(info, index=["train_id", "col3", "col4"])
            pandas_t1 = pd.DataFrame(t1, index=["tid", "col5"])
            trainable = Join(
                pred=[
                    it.main.TrainId == it.info.train_id,
                    it.info.train_id == it.t1.tid,
                ],
                join_type="left",
            )
            pandas_transformed_df = trainable.transform(
                [pandas_main, pandas_info, pandas_t1]
            ).sort("TrainId")
            cls.tgt2datasets = {"pandas": pandas_transformed_df}

    def test_filter_isnan(self):
        pandas_transformed_df = self.tgt2datasets["pandas"]
        self.assertEqual(pandas_transformed_df.shape, (5, 9))
        self.assertEqual(pandas_transformed_df["col1"][2], "CA")
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(pred=[isnan(it.col6)])
            filtered_df = trainable.transform(transformed_df)
            if tgt == "pandas":
                # `None` is considered as `nan` in Pandas
                self.assertEqual(filtered_df.shape, (2, 9), tgt)
                self.assertTrue(all(np.isnan(filtered_df["col6"])), tgt)
            elif tgt.startswith("spark"):
                self.assertEqual(_ensure_pandas(filtered_df).shape, (1, 9), tgt)
                test_list = [row[0] for row in filtered_df.select("col6").collect()]
                self.assertTrue(all((np.isnan(i) for i in test_list if i is not None)))
            else:
                assert False
            if tgt == "spark-with-index":
                self.assertEqual(
                    get_index_name(transformed_df), get_index_name(filtered_df)
                )

    def test_filter_isnotnan(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(pred=[isnotnan(it.col6)])
            filtered_df = trainable.transform(transformed_df)
            if tgt == "pandas":
                self.assertTrue(all(np.logical_not(np.isnan(filtered_df["col6"]))), tgt)
                self.assertEqual(filtered_df.shape, (3, 9), tgt)
            elif tgt.startswith("spark"):
                self.assertEqual(_ensure_pandas(filtered_df).shape, (4, 9), tgt)
                test_list = [row[0] for row in filtered_df.select("col6").collect()]
                self.assertTrue(
                    all((not np.isnan(i) for i in test_list if i is not None))
                )
            else:
                assert False
            if tgt == "spark-with-index":
                self.assertEqual(
                    get_index_name(transformed_df), get_index_name(filtered_df)
                )

    def test_filter_isnull(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(pred=[isnull(it.col6)])
            filtered_df = trainable.transform(transformed_df)
            if tgt == "pandas":
                # `None` is considered as `nan` in Pandas
                self.assertEqual(filtered_df.shape, (2, 9), tgt)
                self.assertTrue(all(np.isnan(filtered_df["col6"])), tgt)
            elif tgt.startswith("spark"):
                self.assertEqual(_ensure_pandas(filtered_df).shape, (1, 9), tgt)
                test_list = [row[0] for row in filtered_df.select("col6").collect()]
                self.assertTrue(all((i is None for i in test_list)))
            else:
                assert False
            if tgt == "spark-with-index":
                self.assertEqual(
                    get_index_name(transformed_df), get_index_name(filtered_df)
                )

    def test_filter_isnotnull(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(pred=[isnotnull(it.col6)])
            filtered_df = trainable.transform(transformed_df)
            if tgt == "pandas":
                # `None` is considered as `nan` in Pandas
                self.assertEqual(filtered_df.shape, (3, 9), tgt)
                self.assertTrue(all(np.logical_not(np.isnan(filtered_df["col6"]))))
            elif tgt.startswith("spark"):
                self.assertEqual(_ensure_pandas(filtered_df).shape, (4, 9), tgt)
                test_list = [row[0] for row in filtered_df.select("col6").collect()]
                self.assertTrue(all((i is not None for i in test_list)))
            else:
                assert False
            if tgt == "spark-with-index":
                self.assertEqual(
                    get_index_name(transformed_df), get_index_name(filtered_df)
                )

    def test_filter_eq(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(pred=[it.col3 == "TX"])
            filtered_df = trainable.transform(transformed_df)
            filtered_df = _ensure_pandas(filtered_df)
            self.assertEqual(filtered_df.shape, (2, 9), tgt)
            self.assertTrue(all(filtered_df["col3"] == "TX"), tgt)

    def test_filter_neq(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(pred=[it.col1 != it["col3"]])
            filtered_df = trainable.transform(transformed_df)
            filtered_df = _ensure_pandas(filtered_df)
            self.assertEqual(filtered_df.shape, (3, 9), tgt)
            self.assertTrue(all(filtered_df["col1"] != filtered_df["col3"]), tgt)

    def test_filter_ge(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(pred=[it["col4"] >= 150])
            filtered_df = trainable.transform(transformed_df)
            filtered_df = _ensure_pandas(filtered_df)
            self.assertEqual(filtered_df.shape, (3, 9), tgt)
            self.assertTrue(all(filtered_df["col4"] >= 150), tgt)

    def test_filter_gt(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(pred=[it["col4"] > 150])
            filtered_df = trainable.transform(transformed_df)
            filtered_df = _ensure_pandas(filtered_df)
            self.assertEqual(filtered_df.shape, (2, 9), tgt)
            self.assertTrue(all(filtered_df["col4"] > 150), tgt)

    def test_filter_le(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(pred=[it["col3"] <= "NY"])
            filtered_df = trainable.transform(transformed_df)
            filtered_df = _ensure_pandas(filtered_df)
            self.assertEqual(filtered_df.shape, (3, 9), tgt)
            self.assertTrue(all(filtered_df["col3"] <= "NY"), tgt)

    def test_filter_lt(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(pred=[it["col2"] < it["TrainId"]])
            filtered_df = trainable.transform(transformed_df)
            filtered_df = _ensure_pandas(filtered_df)
            self.assertEqual(filtered_df.shape, (2, 9), tgt)
            self.assertTrue(all(filtered_df["col2"] < filtered_df["TrainId"]), tgt)

    def test_filter_multiple1(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(pred=[it.col3 == "TX", it["col2"] > 4])
            filtered_df = trainable.transform(transformed_df)
            filtered_df = _ensure_pandas(filtered_df)
            self.assertEqual(filtered_df.shape, (1, 9))
            self.assertTrue(all(filtered_df["col3"] == "TX"), tgt)
            self.assertTrue(all(filtered_df["col2"] > 4), tgt)

    def test_filter_multiple2(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(pred=[it.col5 != "Cold", it.train_id < 4])
            filtered_df = trainable.transform(transformed_df)
            if tgt == "pandas":
                self.assertEqual(filtered_df.shape, (2, 9))
            elif tgt.startswith("spark"):
                # `None != "Cold"` is not true in Spark
                filtered_df = _ensure_pandas(filtered_df)
                self.assertEqual(filtered_df.shape, (1, 9))
            else:
                assert False
            self.assertTrue(all(filtered_df["col5"] != "Cold"), tgt)
            self.assertTrue(all(filtered_df["train_id"] < 4), tgt)

    def test_multiple3(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            trainable = Filter(
                pred=[
                    it["tid"] == it["TrainId"],
                    it["col2"] >= it.train_id,
                    it.col3 == "NY",
                ]
            )
            filtered_df = trainable.transform(transformed_df)
            filtered_df = _ensure_pandas(filtered_df)
            self.assertEqual(filtered_df.shape, (1, 9), tgt)
            self.assertTrue(all(filtered_df["tid"] == filtered_df["TrainId"]), tgt)
            self.assertTrue(all(filtered_df["col2"] >= filtered_df["train_id"]), tgt)
            self.assertTrue(all(filtered_df["col3"] == "NY"), tgt)

    def test_filter_no_col_error(self):
        for tgt, transformed_df in self.tgt2datasets.items():
            with self.assertRaises(ValueError):
                trainable = Filter(pred=[it["TrainId"] < it.col_na])
                _ = trainable.transform(transformed_df)


class TestScan(unittest.TestCase):
    def setUp(self):
        self.go_sales = fetch_go_sales_dataset()

    def test_attribute(self):
        with EnableSchemaValidation():
            trained = Scan(table=it.go_products)
            transformed = trained.transform(self.go_sales)
            self.assertEqual(get_table_name(transformed), "go_products")
            self.assertIs(self.go_sales[3], transformed)

    def test_subscript(self):
        with EnableSchemaValidation():
            trained = Scan(table=it["go_products"])
            transformed = trained.transform(self.go_sales)
            self.assertEqual(get_table_name(transformed), "go_products")
            self.assertIs(self.go_sales[3], transformed)

    def test_error1(self):
        with EnableSchemaValidation():
            trained = Scan(table=it.go_products)
            with self.assertRaisesRegex(ValueError, "invalid X"):
                _ = trained.transform(self.go_sales[3])

    def test_error2(self):
        trained = Scan(table=it.unknown_table)
        with self.assertRaisesRegex(ValueError, "could not find 'unknown_table'"):
            _ = trained.transform(self.go_sales)

    def test_error3(self):
        with self.assertRaisesRegex(ValueError, "expected `it.table_name` or"):
            _ = Scan(table=(it.go_products == 42))


# Testing alias operator for pandas and spark dataframes
class TestAlias(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2datasets = {tgt: fetch_go_sales_dataset(tgt) for tgt in targets}

    def test_alias(self):
        for tgt, datasets in self.tgt2datasets.items():
            trainable = Alias(name="test_alias")
            go_products = datasets[3]
            self.assertEqual(get_table_name(go_products), "go_products")
            transformed_df = trainable.transform(go_products)
            self.assertEqual(get_table_name(transformed_df), "test_alias")
            if tgt == "pandas":
                self.assertTrue(_is_pandas_df(transformed_df))
            elif tgt.startswith("spark"):
                self.assertTrue(_is_spark_df(transformed_df))
                transformed_df = transformed_df.toPandas()
            else:
                assert False
            self.assertEqual(transformed_df.shape, (274, 8))

    def test_alias_name_error(self):
        with self.assertRaises(jsonschema.ValidationError):
            _ = Alias()
        with self.assertRaises(jsonschema.ValidationError):
            _ = Alias(name="")
        with self.assertRaises(jsonschema.ValidationError):
            _ = Alias(name="    ")

    def test_filter_name(self):
        for tgt, datasets in self.tgt2datasets.items():
            go_products = datasets[3]
            trained = Filter(pred=[it["Unit cost"] >= 10])
            transformed = trained.transform(go_products)
            self.assertEqual(get_table_name(transformed), "go_products", tgt)
            if tgt == "spark-with-index":
                self.assertEqual(get_index_name(transformed), "index", tgt)

    def test_map_name(self):
        for tgt, datasets in self.tgt2datasets.items():
            go_products = datasets[3]
            trained = Map(columns={"unit_cost": it["Unit cost"]})
            transformed = trained.transform(go_products)
            self.assertEqual(get_table_name(transformed), "go_products", tgt)
            if tgt == "spark-with-index":
                self.assertEqual(get_index_name(transformed), "index", tgt)

    def test_join_name(self):
        for tgt, datasets in self.tgt2datasets.items():
            trained = Join(
                pred=[it.go_1k["Retailer code"] == it.go_retailers["Retailer code"]],
                name="joined_tables",
            )
            transformed = trained.transform(datasets)
            self.assertEqual(get_table_name(transformed), "joined_tables", tgt)

    def test_groupby_name(self):
        for tgt, datasets in self.tgt2datasets.items():
            go_products = datasets[3]
            trained = GroupBy(by=[it["Product line"]])
            transformed = trained.transform(go_products)
            self.assertEqual(get_table_name(transformed), "go_products", tgt)

    def test_aggregate_name(self):
        for tgt, datasets in self.tgt2datasets.items():
            go_daily_sales = datasets[1]
            group_by = GroupBy(by=[it["Retailer code"]])
            aggregate = Aggregate(columns={"min_quantity": min(it.Quantity)})
            trained = group_by >> aggregate
            transformed = trained.transform(go_daily_sales)
            self.assertEqual(get_table_name(transformed), "go_daily_sales", tgt)


# Testing group_by operator for pandas and spark dataframes
class TestGroupBy(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2datasets = {tgt: fetch_go_sales_dataset(tgt) for tgt in targets}

    def test_groupby(self):
        trainable = GroupBy(by=[it["Product line"]])
        for tgt, datasets in self.tgt2datasets.items():
            go_products = datasets[3]
            assert get_table_name(go_products) == "go_products"
            grouped_df = trainable.transform(go_products)
            if tgt == "pandas":
                self.assertEqual(grouped_df.ngroups, 5, tgt)
            aggregate = Aggregate(columns={"count": count(it["Product line"])})
            df = _ensure_pandas(aggregate.transform(grouped_df))
            self.assertEqual(df.shape, (5, 1), tgt)

    def test_groupby1(self):
        trainable = GroupBy(by=[it["Product line"], it.Product])
        for tgt, datasets in self.tgt2datasets.items():
            go_products = datasets[3]
            assert get_table_name(go_products) == "go_products"
            grouped_df = trainable.transform(go_products)
            if tgt == "pandas":
                self.assertEqual(grouped_df.ngroups, 144, tgt)
            aggregate = Aggregate(columns={"count": count(it["Product line"])})
            df = _ensure_pandas(aggregate.transform(grouped_df))
            self.assertEqual(df.shape, (144, 1), tgt)


# Testing Aggregate operator for both pandas and Spark
class TestAggregate(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2datasets = {tgt: fetch_go_sales_dataset(tgt) for tgt in targets}

    def test_sales_not_grouped_single_col(self):
        pipeline = Scan(table=it.go_daily_sales) >> Aggregate(
            columns={
                "min_method_code": min(it["Order method code"]),
                "max_method_code": max(it["Order method code"]),
                "collect_set('Order method code')": collect_set(
                    it["Order method code"]
                ),
                "mode_method_code": mode(it["Order method code"]),
                "median_method_code": median(it["Order method code"]),
            }
        )
        for tgt, datasets in self.tgt2datasets.items():
            result = pipeline.transform(datasets)
            if tgt.startswith("spark"):
                result = result.toPandas()
            self.assertEqual(result.shape, (1, 5), tgt)
            self.assertEqual(result.loc[0, "min_method_code"], 1, tgt)
            self.assertEqual(result.loc[0, "max_method_code"], 7, tgt)
            self.assertEqual(
                sorted(result.loc[0, "collect_set('Order method code')"]),
                [1, 2, 3, 4, 5, 6, 7],
                tgt,
            )
            self.assertEqual(result.loc[0, "mode_method_code"], 5, tgt)
            self.assertEqual(result.loc[0, "median_method_code"], 5, tgt)

    def test_sales_not_grouped_single_func(self):
        pipeline = Aggregate(
            columns={
                "max_method_code": max(it["Order method code"]),
                "max_method_type": max(it["Order method type"]),
            }
        )
        for tgt, datasets in self.tgt2datasets.items():
            result = pipeline.transform(datasets[2])
            if tgt.startswith("spark"):
                result = result.toPandas()
            self.assertEqual(result.shape, (1, 2), tgt)
            self.assertEqual(result.loc[0, "max_method_code"], 12, tgt)
            self.assertEqual(result.loc[0, "max_method_type"], "Web", tgt)

    def test_sales_multi_col_not_grouped(self):
        pipeline = Aggregate(
            columns={
                "min_method_code": min(it["Order method code"]),
                "max_method_code": max(it["Order method code"]),
                "max_method_type": max(it["Order method type"]),
            }
        )
        for tgt, datasets in self.tgt2datasets.items():
            result = pipeline.transform(datasets[2])
            if tgt.startswith("spark"):
                result = result.toPandas()
            self.assertEqual(result.shape, (1, 3), tgt)
            self.assertEqual(result.loc[0, "min_method_code"], 1, tgt)
            self.assertEqual(result.loc[0, "max_method_code"], 12, tgt)
            self.assertEqual(result.loc[0, "max_method_type"], "Web", tgt)

    def test_sales_onekey_grouped(self):
        pipeline = (
            Scan(table=it.go_daily_sales)
            >> GroupBy(by=[it["Retailer code"]])
            >> Aggregate(
                columns={
                    "retailer_code": it["Retailer code"],
                    "min_method_code": min(it["Order method code"]),
                    "max_method_code": max(it["Order method code"]),
                    "min_quantity": min(it["Quantity"]),
                    "method_codes": collect_set(it["Order method code"]),
                    # Mode is not supported on GroupedData as of now.
                    # "mode_method_code": mode(it["Order method code"]),
                    "median_method_code": median(it["Order method code"]),
                }
            )
        )
        for tgt, datasets in self.tgt2datasets.items():
            result = pipeline.transform(datasets)
            if tgt.startswith("spark"):
                result = result.toPandas()
            self.assertEqual(result.shape, (289, 6))
            row = result[result.retailer_code == 1201]
            self.assertEqual(row.loc[row.index[0], "retailer_code"], 1201, tgt)
            self.assertEqual(row.loc[row.index[0], "min_method_code"], 2, tgt)
            self.assertEqual(row.loc[row.index[0], "max_method_code"], 6, tgt)
            self.assertEqual(row.loc[row.index[0], "min_quantity"], 1, tgt)
            self.assertEqual(
                sorted(row.loc[row.index[0], "method_codes"]), [2, 3, 4, 5, 6], tgt
            )
            # self.assertEqual(result.loc[row.index[0], "mode_method_code"], 5, tgt)
            self.assertEqual(result.loc[row.index[0], "median_method_code"], 5, tgt)
            self.assertEqual(result.index.name, "Retailer code", tgt)

    def test_sales_onekey_grouped_single_col(self):
        pipeline = (
            Scan(table=it.go_daily_sales)
            >> GroupBy(by=[it["Retailer code"]])
            >> Aggregate(
                columns={
                    "min_method_code": min(it["Order method code"]),
                    "max_method_code": max(it["Order method code"]),
                    "method_codes": collect_set(it["Order method code"]),
                    "median_method_code": median(it["Order method code"]),
                }
            )
        )
        for tgt, datasets in self.tgt2datasets.items():
            result = pipeline.transform(datasets)
            if tgt.startswith("spark"):
                result = result.toPandas()
            self.assertEqual(result.shape, (289, 4))
            self.assertEqual(result.loc[1201, "min_method_code"], 2, tgt)
            self.assertEqual(result.loc[1201, "max_method_code"], 6, tgt)
            self.assertEqual(
                sorted(result.loc[1201, "method_codes"]), [2, 3, 4, 5, 6], tgt
            )
            self.assertEqual(result.loc[1201, "median_method_code"], 5, tgt)
            self.assertEqual(result.index.name, "Retailer code", tgt)

    def test_sales_onekey_grouped_single_func(self):
        pipeline = (
            Scan(table=it.go_daily_sales)
            >> GroupBy(by=[it["Retailer code"]])
            >> Aggregate(
                columns={
                    "min_method_code": min(it["Order method code"]),
                    "min_quantity": min(it["Quantity"]),
                }
            )
        )
        for tgt, datasets in self.tgt2datasets.items():
            result = pipeline.transform(datasets)
            if tgt.startswith("spark"):
                result = result.toPandas()
            self.assertEqual(result.shape, (289, 2))
            self.assertEqual(result.loc[1201, "min_method_code"], 2, tgt)
            self.assertEqual(result.loc[1201, "min_quantity"], 1, tgt)
            self.assertEqual(result.index.name, "Retailer code", tgt)

    def test_products_onekey_grouped(self):
        pipeline = (
            Scan(table=it.go_products)
            >> GroupBy(by=[it["Product line"]])
            >> Aggregate(
                columns={
                    "line": first(it["Product line"]),
                    "mean_uc": mean(it["Unit cost"]),
                    "min_up": min(it["Unit price"]),
                    "count_pc": count(it["Product color"]),
                }
            )
        )
        for tgt, datasets in self.tgt2datasets.items():
            result = pipeline.transform(datasets)
            if tgt.startswith("spark"):
                result = result.toPandas()
            self.assertEqual(result.shape, (5, 4))
            row = result[result.line == "Camping Equipment"]
            self.assertEqual(row.loc[row.index[0], "line"], "Camping Equipment", tgt)
            self.assertAlmostEqual(row.loc[row.index[0], "mean_uc"], 89.0, 1, tgt)
            self.assertEqual(row.loc[row.index[0], "min_up"], 2.06, tgt)
            self.assertEqual(row.loc[row.index[0], "count_pc"], 41, tgt)

    def test_sales_twokeys_grouped(self):
        pipeline = (
            Scan(table=it.go_daily_sales)
            >> GroupBy(by=[it["Product number"], it["Retailer code"]])
            >> Aggregate(
                columns={
                    "product": it["Product number"],
                    "retailer": it["Retailer code"],
                    "mean_quantity": mean(it.Quantity),
                    "max_usp": max(it["Unit sale price"]),
                    "count_quantity": count(it.Quantity),
                }
            )
        )
        for tgt, datasets in self.tgt2datasets.items():
            result = pipeline.transform(datasets)
            result = _ensure_pandas(result)
            self.assertEqual(result.shape, (5000, 5))
            row = result[(result["product"] == 70240) & (result["retailer"] == 1205)]
            self.assertEqual(row.loc[row.index[0], "product"], 70240, tgt)
            self.assertEqual(row.loc[row.index[0], "retailer"], 1205, tgt)
            self.assertAlmostEqual(
                row.loc[row.index[0], "mean_quantity"], 48.39, 2, tgt
            )
            self.assertEqual(row.loc[row.index[0], "max_usp"], 122.70, tgt)
            self.assertEqual(row.loc[row.index[0], "count_quantity"], 41, tgt)
            self.assertEqual(
                result.index.names, ["Product number", "Retailer code"], tgt
            )

    def test_products_twokeys_grouped(self):
        pipeline = (
            Scan(table=it.go_products)
            >> GroupBy(by=[it["Product line"], it["Product brand"]])
            >> Aggregate(
                columns={
                    "sum_uc": sum(it["Unit cost"]),
                    "max_uc": max(it["Unit cost"]),
                    "line": first(it["Product line"]),
                    "brand": first(it["Product brand"]),
                }
            )
        )
        for tgt, datasets in self.tgt2datasets.items():
            result = pipeline.transform(datasets)
            if tgt.startswith("spark"):
                result = result.toPandas()
            self.assertEqual(result.shape, (30, 4))
            row = result[
                (result.line == "Camping Equipment") & (result.brand == "Star")
            ]
            self.assertEqual(row.loc[row.index[0], "sum_uc"], 1968.19, tgt)
            self.assertEqual(row.loc[row.index[0], "max_uc"], 490.00, tgt)
            self.assertEqual(row.loc[row.index[0], "line"], "Camping Equipment", tgt)
            self.assertEqual(row.loc[row.index[0], "brand"], "Star", tgt)

    def test_error_unknown_column(self):
        pipeline = (
            Scan(table=it.go_daily_sales)
            >> GroupBy(by=[it["Product number"]])
            >> Aggregate(columns={"mean_quantity": mean(it["Quantity_1"])})
        )
        with self.assertRaises(KeyError):
            _ = pipeline.transform(self.tgt2datasets["pandas"])

    def test_error_columns_not_dict(self):
        pipeline = (
            Scan(table=it.go_daily_sales)
            >> GroupBy(by=[it["Product number"]])
            >> Aggregate(columns=[mean(it["Quantity_1"])])
        )
        with self.assertRaises(ValueError):
            _ = pipeline.transform(self.tgt2datasets["pandas"])

    def test_error_X_not_pandas_or_Spark(self):
        trainable = Aggregate(columns={"mean_quantity": mean(it["Quantity_1"])})
        with self.assertRaises(ValueError):
            _ = trainable.transform(pd.Series([1, 2, 3]))


# Testing join operator for pandas dataframes
class TestJoin(unittest.TestCase):
    def test_init(self):
        _ = Join(pred=[it.main.train_id == it.info.TrainId], join_type="inner")

    # Define pandas dataframes with different structures
    @classmethod
    def setUpClass(cls):
        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2datasets = {
            tgt: {"go_sales": fetch_go_sales_dataset(tgt)} for tgt in targets
        }

        def add_df(name, df):
            cls.tgt2datasets["pandas"][name] = df
            cls.tgt2datasets["spark"][name] = pandas2spark(df)
            cls.tgt2datasets["spark-with-index"][name] = pandas2spark(
                df, with_index=True
            )

        table1 = {
            "train_id": [1, 2, 3, 4, 5],
            "col1": ["NY", "TX", "CA", "NY", "CA"],
            "col2": [0, 1, 1, 0, 1],
        }
        df1 = add_table_name(pd.DataFrame(data=table1), "main")
        add_df("df1", df1)

        table2 = {
            "TrainId": [1, 2, 3],
            "col3": ["USA", "USA", "UK"],
            "col4": [100, 100, 200],
        }
        df2 = add_table_name(pd.DataFrame(data=table2), "info")
        add_df("df2", df2)

        table3 = {
            "tid": [1, 2, 3],
            "col5": ["Warm", "Cold", "Warm"],
        }
        df3 = add_table_name(pd.DataFrame(data=table3), "t1")
        add_df("df3", df3)

        table4 = {
            "TrainId": [1, 2, 3, 4, 5],
            "col1": ["NY", "TX", "CA", "NY", "CA"],
            "col2": [0, 1, 1, 0, 1],
        }
        df4 = add_table_name(pd.DataFrame(data=table4), "main")
        add_df("df4", df4)

        table5 = {
            "TrainId": [1, 2, 3],
            "col3": ["NY", "NY", "CA"],
            "col4": [100, 100, 200],
        }
        df5 = add_table_name(pd.DataFrame(data=table5), "info")
        add_df("df5", df5)

        table6 = {
            "t_id": [2, 3],
            "col6": ["USA", "UK"],
        }
        df6 = add_table_name(pd.DataFrame(data=table6), "t2")
        add_df("df6", df6)

    # Multiple elements in predicate with different key column names
    def test_join_multiple_inner(self):
        trainable = Join(
            pred=[it.main.train_id == it.info.TrainId, it.info.TrainId == it.t1.tid],
            join_type="inner",
        )
        for tgt, datasets in self.tgt2datasets.items():
            df1, df2, df3 = datasets["df1"], datasets["df2"], datasets["df3"]
            transformed_df = trainable.transform([df1, df2, df3])
            transformed_df = _ensure_pandas(transformed_df)
            transformed_df = transformed_df.sort_values(by="train_id").reset_index(
                drop=True
            )
            self.assertEqual(transformed_df.shape, (3, 8), tgt)
            self.assertEqual(transformed_df["col5"][1], "Cold", tgt)

    # Multiple elements in predicate with identical key columns names
    def test_join_multiple_left(self):
        trainable = Join(
            pred=[it.main.TrainId == it.info.TrainId, it.info.TrainId == it.t1.tid],
            join_type="left",
        )
        for tgt, datasets in self.tgt2datasets.items():
            df4, df2, df3 = datasets["df4"], datasets["df2"], datasets["df3"]
            transformed_df = trainable.transform([df4, df2, df3])
            transformed_df = _ensure_pandas(transformed_df)
            transformed_df = transformed_df.sort_values(by="TrainId").reset_index(
                drop=True
            )
            self.assertEqual(transformed_df.shape, (5, 7), tgt)
            self.assertEqual(transformed_df["col3"][2], "UK", tgt)

    # Invert one of the join conditions as compared to the test case: test_join_pandas_multiple_left
    def test_join_multiple_right(self):
        trainable = Join(
            pred=[it.main.train_id == it.info.TrainId, it.t1.tid == it.info.TrainId],
            join_type="right",
        )
        for tgt, datasets in self.tgt2datasets.items():
            df1, df2, df3 = datasets["df1"], datasets["df2"], datasets["df3"]
            transformed_df = trainable.transform([df1, df2, df3])
            transformed_df = _ensure_pandas(transformed_df)
            transformed_df = transformed_df.sort_values(by="TrainId").reset_index(
                drop=True
            )
            self.assertEqual(transformed_df.shape, (3, 8), tgt)
            self.assertEqual(transformed_df["col3"][2], "UK", tgt)

    # Composite key join
    def test_join_composite(self):
        trainable = Join(
            pred=[
                it.t1.tid == it.info.TrainId,
                [it.main.train_id == it.info.TrainId, it.main.col1 == it.info.col3],
            ],
            join_type="left",
        )
        for tgt, datasets in self.tgt2datasets.items():
            df1, df5, df3, df6 = (
                datasets["df1"],
                datasets["df5"],
                datasets["df3"],
                datasets["df6"],
            )
            transformed_df = trainable.transform([df1, df5, df3, df6])
            transformed_df = _ensure_pandas(transformed_df)
            transformed_df = transformed_df.sort_values(by="train_id").reset_index(
                drop=True
            )
            self.assertEqual(transformed_df.shape, (5, 8), tgt)
            self.assertEqual(transformed_df["col3"][2], "CA", tgt)

    # Invert one of the join conditions as compared to the test case: test_join_pandas_composite
    def test_join_composite1(self):
        trainable = Join(
            pred=[
                [it.main.train_id == it.info.TrainId, it.main.col1 == it.info.col3],
                it.t1.tid == it.info.TrainId,
                it.t1.tid == it.t2.t_id,
            ],
            join_type="inner",
        )
        for tgt, datasets in self.tgt2datasets.items():
            df1, df5, df3, df6 = (
                datasets["df1"],
                datasets["df5"],
                datasets["df3"],
                datasets["df6"],
            )
            transformed_df = trainable.transform([df1, df5, df3, df6])
            transformed_df = _ensure_pandas(transformed_df)
            transformed_df = transformed_df.sort_values(by="train_id").reset_index(
                drop=True
            )
            self.assertEqual(transformed_df.shape, (1, 10), tgt)
            self.assertEqual(transformed_df["col4"][0], 200, tgt)

    # Composite key join having conditions involving more than 2 tables
    # This test case execution should throw a ValueError which is handled in the test case itself
    def test_join_composite_error(self):
        with self.assertRaisesRegex(
            ValueError, "info.*main.*inFo.* more than two tables"
        ):
            _ = Join(
                pred=[
                    it.t1.tid == it.info.TrainId,
                    [it.main.train_id == it.info.TrainId, it.main.col1 == it.inFo.col3],
                    it.t1.tid == it.t2.t_id,
                ],
                join_type="inner",
            )

    # Single joining conditions are not chained
    # This test case execution should throw a ValueError which is handled in the test case itself
    def test_join_single_error1(self):
        with self.assertRaisesRegex(ValueError, "t3.*t2.* were used"):
            _ = Join(
                pred=[
                    it.t1.tid == it.info.TrainId,
                    [it.main.train_id == it.info.TrainId, it.main.col1 == it.info.col3],
                    it.t3.tid == it.t2.t_id,
                ],
                join_type="inner",
            )

    def test_join_composite_nochain_error(self):
        with self.assertRaisesRegex(ValueError, "t3.*t2.* were used"):
            _ = Join(
                pred=[
                    it.t1.tid == it.info.TrainId,
                    [it.main.train_id == it.info.TrainId, it.main.col1 == it.info.col3],
                    [it.t3.tid == it.t2.t_id, it.t3.TrainId == it.t2.TrainId],
                ],
                join_type="inner",
            )
            # _ = trainable.transform([self.df1, self.df5, self.df3, self.df6])

    # Composite key join having conditions involving more than 2 tables
    # This test case execution should throw a ValueError which is handled in the test case itself
    def test_join_composite_error2(self):
        with self.assertRaisesRegex(
            ValueError, "main.*info.*Main.*inFo.*more than two"
        ):
            _ = Join(
                pred=[
                    it.t1.tid == it.info.TrainId,
                    [it.main.train_id == it.info.TrainId, it.Main.col1 == it.inFo.col3],
                    it.t1.tid == it.t2.t_id,
                ],
                join_type="inner",
            )

    # A table to be joined not present in input X
    # This test case execution should throw a ValueError which is handled in the test case itself
    def test_join_composite_error3(self):
        for tgt, datasets in self.tgt2datasets.items():
            df5, df3 = datasets["df5"], datasets["df3"]
            with self.assertRaises(ValueError):
                trainable = Join(
                    pred=[
                        it.t1.tid == it.info.TrainId,
                        [
                            it.main.train_id == it.info.TrainId,
                            it.main.col1 == it.info.col3,
                        ],
                    ],
                    join_type="inner",
                )
                _ = trainable.transform([df5, df3])

    # TestCase 1: Go_Sales dataset with different forms of predicate (join conditions)
    def test_join_go_sales1(self):
        for tgt, datasets in self.tgt2datasets.items():
            go_sales = datasets["go_sales"]
            trainable = Join(
                pred=[
                    it.go_daily_sales["Retailer code"]
                    == it["go_retailers"]["Retailer code"]
                ],
                join_type="inner",
            )
            transformed_df = trainable.transform(go_sales)
            order = ["Retailer code", "Product number", "Date"]
            if tgt == "pandas":
                transformed_df = transformed_df.sort_values(by=order).reset_index(
                    drop=True
                )
                self.assertEqual(transformed_df.shape, (149257, 10), tgt)
                self.assertEqual(transformed_df["Country"][4], "France", tgt)
            elif tgt.startswith("spark"):
                self.assertEqual(len(get_columns(transformed_df)), 10, tgt)
                self.assertEqual(transformed_df.count(), 149257, tgt)
                # transformed_df = transformed_df.orderBy(order).collect()
                # self.assertEqual(transformed_df[4]["Country"], "France", tgt)
            else:
                assert False

    # TestCase 2: Go_Sales dataset throws error because of duplicate non-key columns
    def test_join_go_sales2(self):
        for tgt, datasets in self.tgt2datasets.items():
            go_sales = datasets["go_sales"]
            trainable = Join(
                pred=[
                    [
                        it["go_1k"]["Retailer code"]
                        == it.go_daily_sales["Retailer code"],
                        it.go_1k["Product number"]
                        == it["go_daily_sales"]["Product number"],
                    ]
                ],
                join_type="left",
            )
            with self.assertRaises(ValueError):
                _ = trainable.transform(go_sales)

    def test_join_index(self):
        trainable = Join(
            pred=[it.info.idx == it.main.idx, it.info.idx == it.t1.idx],
            join_type="inner",
        )
        df1 = _set_index_name(self.tgt2datasets["pandas"]["df1"], "idx")
        df2 = _set_index_name(self.tgt2datasets["pandas"]["df2"], "idx")
        df3 = _set_index_name(self.tgt2datasets["pandas"]["df3"], "idx")
        for tgt in ["spark-with-index"]:
            if tgt == "spark-with-index":
                df1 = pandas2spark(df1, with_index=True)
                df2 = pandas2spark(df2, with_index=True)
                df3 = pandas2spark(df3, with_index=True)
            transformed_df = trainable.transform([df1, df2, df3])
            transformed_df = _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.index.name, "idx", tgt)
            transformed_df = transformed_df.sort_values(by="TrainId").reset_index(
                drop=True
            )
            self.assertEqual(transformed_df.shape, (3, 8), tgt)
            self.assertEqual(transformed_df["col5"][1], "Cold", tgt)

    def test_join_one_index_right(self):
        trainable = Join(
            pred=[it.info.TrainId == it.main.train_id, it.info.TrainId == it.t1.tid],
            join_type="inner",
        )
        df1 = _set_index(self.tgt2datasets["pandas"]["df1"], "train_id")
        df2 = self.tgt2datasets["pandas"]["df2"]
        df3 = self.tgt2datasets["pandas"]["df3"]
        for tgt in ["spark-with-index"]:
            if tgt == "spark-with-index":
                df1 = pandas2spark(df1, with_index=True)
                df2 = pandas2spark(df2, with_index=True)
                df3 = pandas2spark(df3)
            transformed_df = trainable.transform([df1, df2, df3])
            transformed_df = _ensure_pandas(transformed_df)
            transformed_df = transformed_df.sort_values(by="TrainId").reset_index(
                drop=True
            )
            self.assertEqual(transformed_df.shape, (3, 7))
            self.assertEqual(transformed_df["col5"][1], "Cold")

    def test_join_one_index_left(self):
        trainable = Join(
            pred=[it.main.train_id == it.info.TrainId, it.info.TrainId == it.t1.tid],
            join_type="inner",
        )
        df1 = _set_index(self.tgt2datasets["pandas"]["df1"], "train_id")
        df2 = self.tgt2datasets["pandas"]["df2"]
        df3 = self.tgt2datasets["pandas"]["df3"]
        for tgt in ["spark-with-index"]:
            if tgt == "spark-with-index":
                df1 = pandas2spark(df1, with_index=True)
                df2 = pandas2spark(df2, with_index=True)
                df3 = pandas2spark(df3)
            transformed_df = trainable.transform([df1, df2, df3])
            transformed_df = _ensure_pandas(transformed_df)
            transformed_df = transformed_df.sort_values(by="TrainId").reset_index(
                drop=True
            )
            self.assertEqual(transformed_df.shape, (3, 7))
            self.assertEqual(transformed_df["col5"][1], "Cold")

    def test_join_index_multiple_names(self):
        trainable = Join(
            pred=[it.info.TrainId == it.main.train_id, it.info.TrainId == it.t1.tid],
            join_type="inner",
        )
        df1 = _set_index(self.tgt2datasets["pandas"]["df1"], "train_id")
        df2 = _set_index(self.tgt2datasets["pandas"]["df2"], "TrainId")
        df3 = _set_index(self.tgt2datasets["pandas"]["df3"], "tid")
        for tgt in ["spark-with-index"]:
            if tgt == "spark-with-index":
                df1 = pandas2spark(df1, with_index=True)
                df2 = pandas2spark(df2, with_index=True)
                df3 = pandas2spark(df3, with_index=True)
        transformed_df = trainable.transform([df1, df2, df3])
        transformed_df = _ensure_pandas(transformed_df)
        transformed_df = transformed_df.sort_values(by="TrainId").reset_index(drop=True)
        self.assertEqual(transformed_df.shape, (3, 6))
        self.assertEqual(transformed_df["col5"][1], "Cold")


class TestMap(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2datasets = {
            tgt: {"go_sales": fetch_go_sales_dataset(tgt)} for tgt in targets
        }

        def add_df(name, df):
            cls.tgt2datasets["pandas"][name] = df
            cls.tgt2datasets["spark"][name] = pandas2spark(df)
            cls.tgt2datasets["spark-with-index"][name] = pandas2spark(
                df, with_index=True
            )

        df = pd.DataFrame(
            {
                "gender": ["m", "f", "m", "m", "f"],
                "state": ["NY", "NY", "CA", "NY", "CA"],
                "status": [0, 1, 1, 0, 1],
            }
        )
        add_df("df", df)
        df_date = pd.DataFrame(
            {"date_column": ["2016-05-28", "2016-06-27", "2016-07-26"]}
        )
        add_df("df_date", df_date)
        df_date_alt = pd.DataFrame(
            {"date_column": ["28/05/2016", "27/06/2016", "26/07/2016"]}
        )
        add_df("df_date_alt", df_date_alt)
        df_date_time = pd.DataFrame(
            {
                "date_column": [
                    "2016-01-01 15:16:45",
                    "2016-06-28 12:18:51",
                    "2016-07-28 01:01:01",
                ]
            }
        )
        add_df("df_date_time", df_date_time)
        df_num = pd.DataFrame(
            {
                "height": [3, 4, 6, 3, 5],
                "weight": [30, 50, 170, 40, 130],
                "status": [0, 1, 1, 0, 1],
            }
        )
        add_df("df_num", df_num)
        df_month = pd.DataFrame(
            {
                "month": ["jan", "feb", "mar", "may", "aug"],
            }
        )
        add_df("df_month", df_month)

    def test_init(self):
        gender_map = {"m": "Male", "f": "Female"}
        state_map = {"NY": "New York", "CA": "California"}
        _ = Map(columns=[replace(it.gender, gender_map), replace(it.state, state_map)])

    # The rename column functionality implemented as part of identity function for Map operator
    # does not support explicit identity calls for now.
    def test_transform_identity_map(self):
        trainable = Map(
            columns={
                "new_gender": it.gender,
                "new_status": it["status"],
            }
        )
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(df["gender"][0], transformed_df["new_gender"][0], tgt)
            self.assertEqual(df["status"][3], transformed_df["new_status"][3], tgt)
            self.assertEqual(len(transformed_df.columns), 2, tgt)

    def test_transform_identity_map_implicit_name(self):
        trainable = Map(columns=[identity(it.gender), identity(it["status"])])
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(df["gender"][0], transformed_df["gender"][0], tgt)
            self.assertEqual(df["status"][3], transformed_df["status"][3], tgt)
            self.assertEqual(len(transformed_df.columns), 2, tgt)

    def test_transform_identity_map_passthrough(self):
        trainable = Map(
            columns={
                "new_gender": it.gender,
                "new_status": it["status"],
            },
            remainder="passthrough",
        )
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(df["gender"][0], transformed_df["new_gender"][0])
            self.assertEqual(df["status"][3], transformed_df["new_status"][3])
            self.assertEqual(df["state"][3], transformed_df["state"][3])
            self.assertEqual(len(transformed_df.columns), 3)

    def test_transform_identity_map_error(self):
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            with self.assertRaises(ValueError):
                trainable = Map(columns={"   ": it.gender})
                trained = trainable.fit(df)
                _ = trained.transform(df)
            with self.assertRaises(ValueError):
                trainable = Map(columns={"new_name": it["  "]})
                trained = trainable.fit(df)
                _ = trained.transform(df)
            with self.assertRaises(ValueError):
                trainable = Map(columns=[it.gender])
                trained = trainable.fit(df)
                _ = trained.transform(df)
            with self.assertRaises(ValueError):
                trainable = Map(columns=[it.dummy])
                trained = trainable.fit(df)
                _ = trained.transform(df)

    def test_transform_replace_list_and_remainder(self):
        gender_map = {"m": "Male", "f": "Female"}
        state_map = {"NY": "New York", "CA": "California"}
        trainable = Map(
            columns=[replace(it.gender, gender_map), replace(it.state, state_map)],
            remainder="passthrough",
        )
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (5, 3))
            self.assertEqual(transformed_df["gender"][0], "Male")
            self.assertEqual(transformed_df["state"][0], "New York")
            self.assertEqual(transformed_df["status"][0], 0)

    def test_transform_replace_list(self):
        gender_map = {"m": "Male", "f": "Female"}
        state_map = {"NY": "New York", "CA": "California"}
        trainable = Map(
            columns=[replace(it.gender, gender_map), replace(it.state, state_map)]
        )
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (5, 2))
            self.assertEqual(transformed_df["gender"][0], "Male")
            self.assertEqual(transformed_df["state"][0], "New York")

    def test_transform_replace_map(self):
        gender_map = {"m": "Male", "f": "Female"}
        state_map = {"NY": "New York", "CA": "California"}
        trainable = Map(
            columns={
                "new_gender": replace(it.gender, gender_map),
                "new_state": replace(it.state, state_map),
            }
        )
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (5, 2))
            self.assertEqual(transformed_df["new_gender"][0], "Male")
            self.assertEqual(transformed_df["new_state"][0], "New York")

    def test_transform_dom_list(self):
        trainable = Map(columns=[day_of_month(it.date_column)])
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_date"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df["date_column"][0], 28)
            self.assertEqual(transformed_df["date_column"][1], 27)
            self.assertEqual(transformed_df["date_column"][2], 26)

    def test_transform_dom_fmt_list(self):
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(columns=[day_of_month(it.date_column, "%Y-%m-%d")])
            elif tgt.startswith("spark"):
                trainable = Map(columns=[day_of_month(it.date_column, "y-M-d")])
            else:
                assert False
            df = datasets["df_date"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df["date_column"][0], 28)
            self.assertEqual(transformed_df["date_column"][1], 27)
            self.assertEqual(transformed_df["date_column"][2], 26)

    def test_transform_dom_fmt_map(self):
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(
                    columns={"dom": day_of_month(it.date_column, "%Y-%m-%d")}
                )
            elif tgt.startswith("spark"):
                trainable = Map(columns={"dom": day_of_month(it.date_column, "y-M-d")})
            else:
                assert False
            df = datasets["df_date"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (3, 1))
            self.assertEqual(transformed_df["dom"][0], 28)
            self.assertEqual(transformed_df["dom"][1], 27)
            self.assertEqual(transformed_df["dom"][2], 26)

    def test_transform_dow_list(self):
        trainable = Map(columns=[day_of_week(it.date_column)])
        for tgt, datasets in self.tgt2datasets.items():
            # Spark and Pandas have a different semantics for `day_of_week`
            df = datasets["df_date"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            if tgt == "pandas":
                self.assertEqual(transformed_df["date_column"][0], 5)
                self.assertEqual(transformed_df["date_column"][1], 0)
                self.assertEqual(transformed_df["date_column"][2], 1)
            elif tgt.startswith("spark"):
                df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
                self.assertEqual(transformed_df["date_column"][0], 7)
                self.assertEqual(transformed_df["date_column"][1], 2)
                self.assertEqual(transformed_df["date_column"][2], 3)
            else:
                assert False

    def test_transform_dow_fmt_list(self):
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(columns=[day_of_week(it.date_column, "%Y-%m-%d")])
                df = datasets["df_date"]
                trained = trainable.fit(df)
                transformed_df = trained.transform(df)
                self.assertEqual(transformed_df.shape, (3, 1))
                self.assertEqual(transformed_df["date_column"][0], 5)
                self.assertEqual(transformed_df["date_column"][1], 0)
                self.assertEqual(transformed_df["date_column"][2], 1)
            elif tgt.startswith("spark"):
                trainable = Map(columns=[day_of_week(it.date_column, "y-M-d")])
                df = datasets["df_date"]
                trained = trainable.fit(df)
                transformed_df = trained.transform(df)
                df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
                self.assertEqual(transformed_df.shape, (3, 1))
                self.assertEqual(transformed_df["date_column"][0], 7)
                self.assertEqual(transformed_df["date_column"][1], 2)
                self.assertEqual(transformed_df["date_column"][2], 3)
            else:
                assert False

    def test_transform_dow_fmt_map(self):
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(
                    columns={"dow": day_of_week(it.date_column, "%Y-%m-%d")}
                )
                df = datasets["df_date"]
                trained = trainable.fit(df)
                transformed_df = trained.transform(df)
                df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
                self.assertEqual(transformed_df.shape, (3, 1))
                self.assertEqual(transformed_df["dow"][0], 5)
                self.assertEqual(transformed_df["dow"][1], 0)
                self.assertEqual(transformed_df["dow"][2], 1)
            elif tgt.startswith("spark"):
                trainable = Map(columns={"dow": day_of_week(it.date_column, "y-M-d")})
                df = datasets["df_date"]
                trained = trainable.fit(df)
                transformed_df = trained.transform(df)
                df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
                self.assertEqual(transformed_df.shape, (3, 1))
                self.assertEqual(transformed_df["dow"][0], 7)
                self.assertEqual(transformed_df["dow"][1], 2)
                self.assertEqual(transformed_df["dow"][2], 3)
            else:
                assert False

    def test_transform_doy_list(self):
        trainable = Map(columns=[day_of_year(it.date_column)])
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_date"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df["date_column"][0], 149)
            self.assertEqual(transformed_df["date_column"][1], 179)
            self.assertEqual(transformed_df["date_column"][2], 208)

    def test_transform_doy_fmt_list(self):
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(columns=[day_of_year(it.date_column, "%Y-%m-%d")])
            elif tgt.startswith("spark"):
                trainable = Map(columns=[day_of_year(it.date_column, "y-M-d")])
            else:
                assert False
            df = datasets["df_date"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df["date_column"][0], 149)
            self.assertEqual(transformed_df["date_column"][1], 179)
            self.assertEqual(transformed_df["date_column"][2], 208)

    def test_transform_doy_fmt_map(self):
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(
                    columns={"doy": day_of_year(it.date_column, "%Y-%m-%d")}
                )
            elif tgt.startswith("spark"):
                trainable = Map(columns={"doy": day_of_year(it.date_column, "y-M-d")})
            else:
                assert False
            df = datasets["df_date"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (3, 1))
            self.assertEqual(transformed_df["doy"][0], 149)
            self.assertEqual(transformed_df["doy"][1], 179)
            self.assertEqual(transformed_df["doy"][2], 208)

    def test_transform_hour_list(self):
        trainable = Map(columns=[hour(it.date_column)])
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_date_time"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df["date_column"][0], 15)
            self.assertEqual(transformed_df["date_column"][1], 12)
            self.assertEqual(transformed_df["date_column"][2], 1)

    def test_transform_hour_fmt_list(self):
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(columns=[hour(it.date_column, "%Y-%m-%d %H:%M:%S")])
            elif tgt.startswith("spark"):
                trainable = Map(columns=[hour(it.date_column, "y-M-d HH:mm:ss")])
            else:
                assert False
            df = datasets["df_date_time"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df["date_column"][0], 15)
            self.assertEqual(transformed_df["date_column"][1], 12)
            self.assertEqual(transformed_df["date_column"][2], 1)

    def test_transform_hour_fmt_map(self):
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(
                    columns={"hour": hour(it.date_column, "%Y-%m-%d %H:%M:%S")}
                )
            elif tgt.startswith("spark"):
                trainable = Map(
                    columns={"hour": hour(it.date_column, "y-M-d HH:mm:ss")}
                )
            else:
                assert False
            df = datasets["df_date_time"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (3, 1))
            self.assertEqual(transformed_df["hour"][0], 15)
            self.assertEqual(transformed_df["hour"][1], 12)
            self.assertEqual(transformed_df["hour"][2], 1)

    def test_transform_minute_list(self):
        trainable = Map(columns=[minute(it.date_column)])
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_date_time"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df["date_column"][0], 16)
            self.assertEqual(transformed_df["date_column"][1], 18)
            self.assertEqual(transformed_df["date_column"][2], 1)

    def test_transform_minute_fmt_list(self):
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(columns=[minute(it.date_column, "%Y-%m-%d %H:%M:%S")])
            elif tgt.startswith("spark"):
                trainable = Map(columns=[minute(it.date_column, "y-M-d HH:mm:ss")])
            else:
                assert False
            df = datasets["df_date_time"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df["date_column"][0], 16)
            self.assertEqual(transformed_df["date_column"][1], 18)
            self.assertEqual(transformed_df["date_column"][2], 1)

    def test_transform_minute_fmt_map(self):
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(
                    columns={"minute": minute(it.date_column, "%Y-%m-%d %H:%M:%S")}
                )
            elif tgt.startswith("spark"):
                trainable = Map(
                    columns={"minute": minute(it.date_column, "y-M-d HH:mm:ss")}
                )
            else:
                assert False
            df = datasets["df_date_time"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (3, 1))
            self.assertEqual(transformed_df["minute"][0], 16)
            self.assertEqual(transformed_df["minute"][1], 18)
            self.assertEqual(transformed_df["minute"][2], 1)

    def test_transform_month_list(self):
        trainable = Map(columns=[month(it.date_column)])
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_date_time"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df["date_column"][0], 1)
            self.assertEqual(transformed_df["date_column"][1], 6)
            self.assertEqual(transformed_df["date_column"][2], 7)

    def test_transform_month_fmt_list(self):
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(columns=[month(it.date_column, "%Y-%m-%d %H:%M:%S")])
            elif tgt.startswith("spark"):
                trainable = Map(columns=[month(it.date_column, "y-M-d HH:mm:ss")])
            else:
                assert False
            df = datasets["df_date_time"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df["date_column"][0], 1)
            self.assertEqual(transformed_df["date_column"][1], 6)
            self.assertEqual(transformed_df["date_column"][2], 7)

    def test_transform_month_fmt_map(self):
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(
                    columns={"month": month(it.date_column, "%Y-%m-%d %H:%M:%S")}
                )
            elif tgt.startswith("spark"):
                trainable = Map(
                    columns={"month": month(it.date_column, "y-M-d HH:mm:ss")}
                )
            else:
                assert False
            df = datasets["df_date_time"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (3, 1))
            self.assertEqual(transformed_df["month"][0], 1)
            self.assertEqual(transformed_df["month"][1], 6)
            self.assertEqual(transformed_df["month"][2], 7)

    def test_not_expression(self):
        with EnableSchemaValidation():
            with self.assertRaises(jsonschema.ValidationError):
                _ = Map(columns=[123, "hello"])

    def test_pandas_with_hyperopt(self):
        from sklearn.datasets import load_iris

        X, y = load_iris(return_X_y=True)
        gender_map = {"m": "Male", "f": "Female"}
        state_map = {"NY": "New York", "CA": "California"}
        map_replace = Map(
            columns=[replace(it.gender, gender_map), replace(it.state, state_map)],
            remainder="drop",
        )
        pipeline = (
            Relational(
                operator=(Scan(table=it.main) & Scan(table=it.delay)) >> map_replace
            )
            >> LogisticRegression()
        )
        opt = Hyperopt(estimator=pipeline, cv=3, max_evals=5)
        trained = opt.fit(X, y)
        _ = trained

    def test_string_indexer_map(self):
        trainable = Map(columns={"c": string_indexer(it.date_column)})
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_date_time"]
            trained = trainable.fit(df)
            with self.assertRaises(ValueError):
                _ = trained.transform(df)

    def test_pands_with_hyperopt2(self):
        from lale.expressions import count, it, max, mean, min, sum, variance

        wrap_imported_operators()
        scan = Scan(table=it["main"])
        scan_0 = Scan(table=it["customers"])
        join = Join(
            pred=[
                (
                    it["main"]["group_customer_id"]
                    == it["customers"]["group_customer_id"]
                )
            ]
        )
        map = Map(
            columns={
                "[main](group_customer_id)[customers]|number_children|identity": it[
                    "number_children"
                ],
                "[main](group_customer_id)[customers]|name|identity": it["name"],
                "[main](group_customer_id)[customers]|income|identity": it["income"],
                "[main](group_customer_id)[customers]|address|identity": it["address"],
                "[main](group_customer_id)[customers]|age|identity": it["age"],
            },
            remainder="drop",
        )
        pipeline_4 = join >> map
        scan_1 = Scan(table=it["purchase"])
        join_0 = Join(
            pred=[(it["main"]["group_id"] == it["purchase"]["group_id"])],
            join_limit=50.0,
        )
        aggregate = Aggregate(
            columns={
                "[main](group_id)[purchase]|price|variance": variance(it["price"]),
                "[main](group_id)[purchase]|time|sum": sum(it["time"]),
                "[main](group_id)[purchase]|time|mean": mean(it["time"]),
                "[main](group_id)[purchase]|time|min": min(it["time"]),
                "[main](group_id)[purchase]|price|sum": sum(it["price"]),
                "[main](group_id)[purchase]|price|count": count(it["price"]),
                "[main](group_id)[purchase]|price|mean": mean(it["price"]),
                "[main](group_id)[purchase]|price|min": min(it["price"]),
                "[main](group_id)[purchase]|price|max": max(it["price"]),
                "[main](group_id)[purchase]|time|max": max(it["time"]),
                "[main](group_id)[purchase]|time|variance": variance(it["time"]),
            },
            group_by=it["row_id"],
        )
        pipeline_5 = join_0 >> aggregate
        map_0 = Map(
            columns={
                "[main]|group_customer_id|identity": it["group_customer_id"],
                "[main]|transaction_id|identity": it["transaction_id"],
                "[main]|group_id|identity": it["group_id"],
                "[main]|comments|identity": it["comments"],
                "[main]|id|identity": it["id"],
                "prefix_0_id": it["prefix_0_id"],
                "next_purchase": it["next_purchase"],
                "[main]|time|identity": it["time"],
            },
            remainder="drop",
        )
        scan_2 = Scan(table=it["transactions"])
        scan_3 = Scan(table=it["products"])
        join_1 = Join(
            pred=[
                (it["main"]["transaction_id"] == it["transactions"]["transaction_id"]),
                (it["transactions"]["product_id"] == it["products"]["product_id"]),
            ]
        )
        map_1 = Map(
            columns={
                "[main](transaction_id)[transactions](product_id)[products]|price|identity": it[
                    "price"
                ],
                "[main](transaction_id)[transactions](product_id)[products]|type|identity": it[
                    "type"
                ],
            },
            remainder="drop",
        )
        pipeline_6 = join_1 >> map_1
        join_2 = Join(
            pred=[
                (it["main"]["transaction_id"] == it["transactions"]["transaction_id"])
            ]
        )
        map_2 = Map(
            columns={
                "[main](transaction_id)[transactions]|description|identity": it[
                    "description"
                ],
                "[main](transaction_id)[transactions]|product_id|identity": it[
                    "product_id"
                ],
            },
            remainder="drop",
        )
        pipeline_7 = join_2 >> map_2
        map_3 = Map(
            columns=[
                string_indexer(it["[main]|comments|identity"]),
                string_indexer(
                    it["[main](transaction_id)[transactions]|description|identity"]
                ),
                string_indexer(
                    it[
                        "[main](transaction_id)[transactions](product_id)[products]|type|identity"
                    ]
                ),
                string_indexer(
                    it["[main](group_customer_id)[customers]|name|identity"]
                ),
                string_indexer(
                    it["[main](group_customer_id)[customers]|address|identity"]
                ),
            ]
        )
        pipeline_8 = ConcatFeatures() >> map_3
        relational = Relational(
            operator=lale.operators.make_pipeline_graph(
                steps=[
                    scan,
                    scan_0,
                    pipeline_4,
                    scan_1,
                    pipeline_5,
                    map_0,
                    scan_2,
                    scan_3,
                    pipeline_6,
                    pipeline_7,
                    pipeline_8,
                ],
                edges=[
                    (scan, pipeline_4),
                    (scan, pipeline_5),
                    (scan, map_0),
                    (scan, pipeline_6),
                    (scan, pipeline_7),
                    (scan_0, pipeline_4),
                    (pipeline_4, pipeline_8),
                    (scan_1, pipeline_5),
                    (pipeline_5, pipeline_8),
                    (map_0, pipeline_8),
                    (scan_2, pipeline_6),
                    (scan_2, pipeline_7),
                    (scan_3, pipeline_6),
                    (pipeline_6, pipeline_8),
                    (pipeline_7, pipeline_8),
                ],
            )
        )
        pipeline = relational >> (KNeighborsClassifier | LogisticRegression)
        from sklearn.datasets import load_iris

        X, y = load_iris(return_X_y=True)

        opt = Hyperopt(estimator=pipeline, max_evals=2)
        opt.fit(X, y)

    def test_transform_ratio_map(self):
        trainable = Map(columns={"ratio_h_w": it.height / it.weight})
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_num"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (5, 1))
            self.assertEqual(transformed_df["ratio_h_w"][0], 0.1)

    def test_transform_ratio_map_subscript(self):
        trainable = Map(columns={"ratio_h_w": it["height"] / it.weight})
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_num"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (5, 1))
            self.assertEqual(transformed_df["ratio_h_w"][0], 0.1)

    def test_transform_ratio_map_list(self):
        trainable = Map(columns=[it.height / it.weight])
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_num"]
            trained = trainable.fit(df)
            with self.assertRaises(ValueError):
                _ = trained.transform(df)

    def test_transform_subtract_map(self):
        trainable = Map(columns={"subtract_h_w": it.height - it.weight})
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_num"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (5, 1))
            self.assertEqual(transformed_df["subtract_h_w"][0], -27)

    def test_transform_subtract_map_subscript(self):
        trainable = Map(columns={"subtract_h_w": it["height"] - it.weight})
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_num"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (5, 1))
            self.assertEqual(transformed_df["subtract_h_w"][0], -27)

    def test_transform_subtract_map_list(self):
        trainable = Map(columns=[it.height - it.weight])
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_num"]
            trained = trainable.fit(df)
            with self.assertRaises(ValueError):
                _ = trained.transform(df)

    def test_transform_binops(self):
        trainable = Map(
            columns={
                "add_h_w": it["height"] + it.weight,
                "add_h_2": it["height"] + 2,
                "sub_h_w": it["height"] - it.weight,
                "sub_h_2": it["height"] - 2,
                "mul_h_w": it["height"] * it.weight,
                "mul_h_2": it["height"] * 2,
                "div_h_w": it["height"] / it.weight,
                "div_h_2": it["height"] / 2,
                "floor_div_h_w": it["height"] // it.weight,
                "floor_div_h_2": it["height"] // 2,
                "mod_h_w": it["height"] % it.weight,
                "mod_h_2": it["height"] % 2,
                "pow_h_w": it["height"] ** it.weight,
                "pow_h_2": it["height"] ** 2,
            }
        )
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_num"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (5, 14))
            self.assertEqual(
                transformed_df["add_h_w"][1], df["height"][1] + df["weight"][1]
            )
            self.assertEqual(transformed_df["add_h_2"][1], df["height"][1] + 2)
            self.assertEqual(
                transformed_df["sub_h_w"][1], df["height"][1] - df["weight"][1]
            )
            self.assertEqual(transformed_df["sub_h_2"][1], df["height"][1] - 2)
            self.assertEqual(
                transformed_df["mul_h_w"][1], df["height"][1] * df["weight"][1]
            )
            self.assertEqual(transformed_df["mul_h_2"][1], df["height"][1] * 2)
            self.assertEqual(
                transformed_df["div_h_w"][1], df["height"][1] / df["weight"][1]
            )
            self.assertEqual(transformed_df["div_h_2"][1], df["height"][1] / 2)
            self.assertEqual(
                transformed_df["floor_div_h_w"][1], df["height"][1] // df["weight"][1]
            )
            self.assertEqual(transformed_df["floor_div_h_2"][1], df["height"][1] // 2)
            self.assertEqual(
                transformed_df["mod_h_w"][1], df["height"][1] % df["weight"][1]
            )
            self.assertEqual(transformed_df["mod_h_2"][1], df["height"][1] % 2)
            if tgt == "pandas":
                self.assertEqual(
                    transformed_df["pow_h_w"][1], df["height"][1] ** df["weight"][1]
                )
            elif tgt.startswith("spark"):
                # Spark and Pandas have a different semantics for large numbers
                self.assertEqual(transformed_df["pow_h_w"][1], 4**50)
            else:
                assert False
            self.assertEqual(transformed_df["pow_h_2"][1], df["height"][1] ** 2)

    def test_transform_arithmetic_expression(self):
        trainable = Map(columns={"expr": (it["height"] + it.weight * 10) / 2})
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_num"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df.shape, (5, 1))
            self.assertEqual(
                transformed_df["expr"][2], (df["height"][2] + df["weight"][2] * 10) / 2
            )

    def test_transform_nested_expressions(self):
        month_map = {
            "jan": "2021-01-01",
            "feb": "2021-02-01",
            "mar": "2021-03-01",
            "arp": "2021-04-01",
            "may": "2021-05-01",
            "jun": "2021-06-01",
            "jul": "2021-07-01",
            "aug": "2021-08-01",
            "sep": "2021-09-01",
            "oct": "2021-10-01",
            "nov": "2021-11-01",
            "dec": "2021-12-01",
        }
        for tgt, datasets in self.tgt2datasets.items():
            if tgt == "pandas":
                trainable = Map(
                    columns={
                        "date": replace(it.month, month_map),
                        "month_id": month(replace(it.month, month_map), "%Y-%m-%d"),
                        "next_month_id": identity(
                            month(replace(it.month, month_map), "%Y-%m-%d") % 12 + 1
                        ),
                    }
                )
            elif tgt.startswith("spark"):
                trainable = Map(
                    columns={
                        "date": replace(it.month, month_map),
                        "month_id": month(replace(it.month, month_map), "y-M-d"),
                        "next_month_id": identity(
                            month(replace(it.month, month_map), "y-M-d") % 12 + 1
                        ),
                    }
                )
            else:
                assert False

            df = datasets["df_month"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertEqual(transformed_df["date"][0], "2021-01-01")
            self.assertEqual(transformed_df["date"][1], "2021-02-01")
            self.assertEqual(transformed_df["month_id"][2], 3)
            self.assertEqual(transformed_df["month_id"][3], 5)
            self.assertEqual(transformed_df["next_month_id"][0], 2)
            self.assertEqual(transformed_df["next_month_id"][3], 6)
            self.assertEqual(transformed_df["next_month_id"][4], 9)

    def test_replace_unknown_identity(self):
        pipeline = Scan(table=it.go_products) >> Map(
            columns={
                "prod": it["Product number"],
                "line": replace(
                    it["Product line"],
                    {"Camping Equipment": "C", "Personal Accessories": "P"},
                ),
            }
        )
        for tgt, datasets in self.tgt2datasets.items():
            result = pipeline.transform(datasets["go_sales"])
            result = _ensure_pandas(result)
            self.assertEqual(result.shape, (274, 2))
            self.assertEqual(result.loc[0, "prod"], 1110, tgt)
            self.assertEqual(result.loc[0, "line"], "C", tgt)
            self.assertEqual(result.loc[117, "prod"], 101110, tgt)
            self.assertEqual(result.loc[117, "line"], "Golf Equipment", tgt)
            self.assertEqual(result.loc[273, "prod"], 154150, tgt)
            self.assertEqual(result.loc[273, "line"], "P", tgt)

    def test_replace_unknown_encoded(self):
        pipeline = Scan(table=it.go_products) >> Map(
            columns={
                "prod": it["Product number"],
                "line": replace(
                    it["Product line"],
                    {"Camping Equipment": "C", "Personal Accessories": "P"},
                    handle_unknown="use_encoded_value",
                    unknown_value="U",
                ),
            }
        )
        for tgt, datasets in self.tgt2datasets.items():
            result = pipeline.transform(datasets["go_sales"])
            result = _ensure_pandas(result)
            self.assertEqual(result.shape, (274, 2))
            self.assertEqual(result.loc[0, "prod"], 1110, tgt)
            self.assertEqual(result.loc[0, "line"], "C", tgt)
            self.assertEqual(result.loc[117, "prod"], 101110, tgt)
            self.assertEqual(result.loc[117, "line"], "U", tgt)
            self.assertEqual(result.loc[273, "prod"], 154150, tgt)
            self.assertEqual(result.loc[273, "line"], "P", tgt)

    def test_dynamic_rename(self):
        def expr(X):
            return {("new_" + c): it[c] for c in X.columns}

        pipeline = Scan(table=it.go_products) >> Map(columns=expr)
        for tgt, datasets in self.tgt2datasets.items():
            datasets = datasets["go_sales"]
            result = pipeline.fit(datasets).transform(datasets)
            result = _ensure_pandas(result)
            for c in result.columns:
                self.assertRegex(c, "new_.*")

    def test_dynamic_rename_lambda(self):
        pipeline = Scan(table=it.go_products) >> Map(
            columns=lambda X: {("new_" + c): it[c] for c in X.columns}
        )
        for tgt, datasets in self.tgt2datasets.items():
            datasets = datasets["go_sales"]
            result = pipeline.fit(datasets).transform(datasets)
            result = _ensure_pandas(result)
            for c in result.columns:
                self.assertRegex(c, "new_.*")

    def _get_col_schemas(self, cols, X):
        from lale.datasets import data_schemas

        props = {}
        s = data_schemas.to_schema(X)
        if s is not None:
            inner = s.get("items", {})
            if inner is not None and isinstance(inner, dict):
                col_pairs = inner.get("items", [])
                if col_pairs is not None and isinstance(col_pairs, list):
                    for cp in col_pairs:
                        d = cp.get("description", None)
                        if d is not None and isinstance(d, str):
                            props[d] = cp
        for k in cols:
            if k not in props:
                props[k] = None
        return props

    def test_dynamic_schema_num(self):
        from lale import type_checking

        def expr(X):
            ret = {}
            schemas = self._get_col_schemas(X.columns, X)
            for c, s in schemas.items():

                if s is None:
                    ret["unknown_" + c] = it[c]
                elif type_checking.is_subschema(s, {"type": "number"}):
                    ret["num_" + c] = it[c]
                    ret["shifted_" + c] = it[c] + 5
                else:
                    ret["other_" + c] = it[c]
            return ret

        pipeline = Scan(table=it.go_products) >> Map(columns=expr)
        for tgt, datasets in self.tgt2datasets.items():
            datasets = datasets["go_sales"]
            result = pipeline.fit(datasets).transform(datasets)
            result = _ensure_pandas(result)
            self.assertIn("num_Product number", result.columns)
            self.assertIn("shifted_Product number", result.columns)
            self.assertIn("other_Product line", result.columns)
            self.assertEqual(
                result["num_Product number"][0] + 5, result["shifted_Product number"][0]
            )

    def test_dynamic_categorical(self):
        from lale.lib.rasl import categorical

        def expr(X):
            ret = {}
            cats = categorical()(X)
            for c in X.columns:
                if c in cats:
                    ret["cat_" + c] = it[c]
                else:
                    ret["other_" + c] = it[c]
            return ret

        pipeline = Scan(table=it.go_products) >> Map(columns=expr)
        for tgt, datasets in self.tgt2datasets.items():
            datasets = datasets["go_sales"]
            result = pipeline.fit(datasets).transform(datasets)
            result = _ensure_pandas(result)
            self.assertIn("cat_Product line", result.columns)

    def test_dynamic_lambda_categorical_drop(self):
        from lale.lib.rasl import categorical

        pipeline = Scan(table=it.go_products) >> Map(
            columns=lambda X: {c: it[c] for c in categorical()(_ensure_pandas(X))}
        )
        for tgt, datasets in self.tgt2datasets.items():
            datasets = datasets["go_sales"]
            result = pipeline.fit(datasets).transform(datasets)
            result = _ensure_pandas(result)
            self.assertEqual(len(result.columns), 1)
            self.assertIn("Product line", result.columns)

    def test_static_trained(self):
        op = Map(columns=[it.col])
        self.assertIsInstance(op, lale.operators.TrainedOperator)

    def test_dynamic_trainable(self):
        op = Map(columns=lambda X: [it.col])
        self.assertIsInstance(op, lale.operators.TrainableOperator)
        self.assertNotIsInstance(op, lale.operators.TrainedOperator)

        pipeline = Scan(table=it.go_products) >> op
        pd = self.tgt2datasets["pandas"]["go_sales"]
        trained = pipeline.fit(pd)
        trained_map = trained.steps_list()[1]
        self.assertIsInstance(trained_map, Map)  # type: ignore
        self.assertIsInstance(trained_map, lale.operators.TrainedOperator)

    def test_project(self):
        from lale.lib.lale import Project

        pipeline = Scan(table=it.go_products) >> Project(columns={"type": "number"})
        for tgt, datasets in self.tgt2datasets.items():
            datasets = datasets["go_sales"]
            result = pipeline.fit(datasets).transform(datasets)
            result = _ensure_pandas(result)
            self.assertIn("Product number", result.columns)
            self.assertNotIn("Product line", result.columns)

    def assertSeriesEqual(self, first, second, msg=None):
        self.assertIsInstance(first, pd.Series, msg)
        self.assertIsInstance(second, pd.Series, msg)
        self.assertEqual(first.shape, second.shape, msg)
        self.assertEqual(list(first), list(second), msg)

    def test_transform_compare_ops(self):
        trained = Map(
            columns={
                "height<=5": it.height <= 5,
                "int(height<=5)": astype("int", it.height <= 5),
                "4==height": 4 == it.height,
                "height*10==weight": it.height * 10 == it.weight,
                "height>3&<=5": (it.height > 3) & (it.height <= 5),
                "height<=3|>5": (it.height <= 3) | (it.height > 5),
            }
        )
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df_num"]
            transformed_df = trained.transform(df)
            df, transformed_df = _ensure_pandas(df), _ensure_pandas(transformed_df)
            self.assertSeriesEqual(transformed_df["height<=5"], df["height"] <= 5, tgt)
            self.assertSeriesEqual(
                transformed_df["int(height<=5)"], (df["height"] <= 5).astype(int), tgt
            )
            self.assertSeriesEqual(transformed_df["4==height"], 4 == df["height"], tgt)
            self.assertSeriesEqual(
                transformed_df["height*10==weight"], df["height"] * 10 == df.weight, tgt
            )
            self.assertSeriesEqual(
                transformed_df["height>3&<=5"], (df["height"] > 3) & (df["height"] <= 5)
            )
            self.assertSeriesEqual(
                transformed_df["height<=3|>5"], (df["height"] <= 3) | (df["height"] > 5)
            )


class TestRelationalOperator(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        from sklearn.datasets import load_iris
        from sklearn.model_selection import train_test_split

        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2datasets = {tgt: {} for tgt in targets}

        def add_df(name, df):
            cls.tgt2datasets["pandas"][name] = df
            cls.tgt2datasets["spark"][name] = pandas2spark(df)
            cls.tgt2datasets["spark-with-index"][name] = pandas2spark(
                df, with_index=True
            )

        X, y = load_iris(as_frame=True, return_X_y=True)
        X_train, X_test, y_train, y_test = train_test_split(X, y)
        add_df("X_train", X_train)
        add_df("X_test", X_test)
        add_df("y_train", y_train)
        add_df("y_test", y_test)

    def test_fit_transform(self):
        relational = Relational(
            operator=(Scan(table=it.main) & Scan(table=it.delay))
            >> Join(
                pred=[
                    it.main.TrainId == it.delay.TrainId,
                    it.main["Arrival time"] >= it.delay.TimeStamp,
                ]
            )
            >> Aggregate(columns=[count(it.Delay)], group_by=it.MessageId)
        )
        for tgt, datasets in self.tgt2datasets.items():
            X_train, X_test, y_train = (
                datasets["X_train"],
                datasets["X_test"],
                datasets["y_train"],
            )
            trained_relational = relational.fit(X_train, y_train)
            _ = trained_relational.transform(X_test)

    def test_fit_error(self):
        relational = Relational(
            operator=(Scan(table=it.main) & Scan(table=it.delay))
            >> Join(
                pred=[
                    it.main.TrainId == it.delay.TrainId,
                    it.main["Arrival time"] >= it.delay.TimeStamp,
                ]
            )
            >> Aggregate(columns=[count(it.Delay)], group_by=it.MessageId)
        )
        for tgt, datasets in self.tgt2datasets.items():
            X_train, y_train = datasets["X_train"], datasets["y_train"]
            with self.assertRaises(ValueError):
                _ = relational.fit([X_train], y_train)

    def test_transform_error(self):
        relational = Relational(
            operator=(Scan(table=it.main) & Scan(table=it.delay))
            >> Join(
                pred=[
                    it.main.TrainId == it.delay.TrainId,
                    it.main["Arrival time"] >= it.delay.TimeStamp,
                ]
            )
            >> Aggregate(columns=[count(it.Delay)], group_by=it.MessageId)
        )
        for tgt, datasets in self.tgt2datasets.items():
            X_train, X_test, y_train = (
                datasets["X_train"],
                datasets["X_test"],
                datasets["y_train"],
            )
            trained_relational = relational.fit(X_train, y_train)
            with self.assertRaises(ValueError):
                _ = trained_relational.transform([X_test])

    def test_fit_transform_in_pipeline(self):
        relational = Relational(
            operator=(Scan(table=it.main) & Scan(table=it.delay))
            >> Join(
                pred=[
                    it.main.TrainId == it.delay.TrainId,
                    it.main["Arrival time"] >= it.delay.TimeStamp,
                ]
            )
            >> Aggregate(columns=[count(it.Delay)], group_by=it.MessageId)
        )
        pipeline = relational >> LogisticRegression()
        for tgt, datasets in self.tgt2datasets.items():
            X_train, X_test, y_train = (
                datasets["X_train"],
                datasets["X_test"],
                datasets["y_train"],
            )
            if tgt == "pandas":
                trained_pipeline = pipeline.fit(X_train, y_train)
                _ = trained_pipeline.predict(X_test)
            elif tgt.startswith("spark"):
                # LogisticRegression is not implemented on Spark
                pass
            else:
                assert False


class TestOrderBy(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2datasets = {tgt: {} for tgt in targets}

        def add_df(name, df):
            cls.tgt2datasets["pandas"][name] = df
            cls.tgt2datasets["spark"][name] = pandas2spark(df)
            cls.tgt2datasets["spark-with-index"][name] = pandas2spark(
                df, with_index=True
            )

        df = pd.DataFrame(
            {
                "gender": ["m", "f", "m", "m", "f"],
                "state": ["NY", "NY", "CA", "NY", "CA"],
                "status": [0, 1, 1, 0, 1],
            }
        )
        add_df("df", df)

    def test_order_attr1(self):
        trainable = OrderBy(by=it.status)
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            if tgt == "spark-with-index":
                self.assertEqual(get_index_name(transformed_df), get_index_name(df))
            transformed_df = _ensure_pandas(transformed_df)
            self.assertTrue((transformed_df["status"]).is_monotonic_increasing)

    def test_order_attr1_asc(self):
        trainable = OrderBy(by=asc(it.status))
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            if tgt == "spark-with-index":
                self.assertEqual(get_index_name(transformed_df), get_index_name(df))
            transformed_df = _ensure_pandas(transformed_df)
            self.assertTrue((transformed_df["status"]).is_monotonic_increasing)

    def test_order_attr1_desc(self):
        trainable = OrderBy(by=desc(it.status))
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            transformed_df = _ensure_pandas(transformed_df)
            self.assertTrue((transformed_df["status"]).is_monotonic_decreasing)

    def test_order_str1_desc(self):
        trainable = OrderBy(by=desc("gender"))
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            transformed_df = _ensure_pandas(transformed_df)
            self.assertTrue((transformed_df["gender"]).is_monotonic_decreasing)

    def test_order_multiple(self):
        trainable = OrderBy(by=[it.gender, desc(it.status)])
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            transformed_df = _ensure_pandas(transformed_df)
            expected_result = pd.DataFrame(
                data={
                    "gender": ["f", "f", "m", "m", "m"],
                    "state": ["NY", "CA", "CA", "NY", "NY"],
                    "status": [1, 1, 1, 0, 0],
                }
            )
            if tgt == "pandas" or tgt == "spark-with-index":
                self.assertEqual(list(transformed_df.index), [1, 4, 2, 0, 3])
            self.assertTrue((transformed_df["gender"]).is_monotonic_increasing)
            self.assertTrue(
                transformed_df.reset_index(drop=True).equals(expected_result)
            )

    def test_str1(self):
        trainable = OrderBy(by="gender")
        for tgt, datasets in self.tgt2datasets.items():
            df = datasets["df"]
            trained = trainable.fit(df)
            transformed_df = trained.transform(df)
            transformed_df = _ensure_pandas(transformed_df)
            self.assertTrue((transformed_df["gender"]).is_monotonic_increasing)


class TestSplitXy(unittest.TestCase):
    @classmethod
    def setUp(cls):
        data = load_iris()
        X, y = data.data, data.target
        X_train, X_test, y_train, y_test = train_test_split(
            pd.DataFrame(X), pd.DataFrame(y)
        )
        combined_df = pd.concat([X_train, y_train], axis=1)
        combined_df.columns = [
            "sepal_length",
            "sepal_width",
            "petal_length",
            "petal_width",
            "class",
        ]
        spark_df = pandas2spark(combined_df)
        cls.tgt2datasets = {
            "pandas": combined_df,
            "spark": spark_df,
            "spark-with-index": SparkDataFrameWithIndex(spark_df),
        }

    def test_split_transform(self):
        for _, df in self.tgt2datasets.items():
            trainable = SplitXy(label_name="class") >> Convert(astype="pandas") >> PCA()
            trained = trainable.fit(df)
            _ = trained.transform(df)

    def test_split_predict(self):
        for _, df in self.tgt2datasets.items():
            trainable = (
                SplitXy(label_name="class")
                >> Convert(astype="pandas")
                >> PCA()
                >> LogisticRegression(random_state=42)
            )
            trained = trainable.fit(df)
            _ = trained.predict(df)


class TestTrainTestSplit(unittest.TestCase):
    # Get go_sales dataset in pandas and spark dataframes
    def setUp(self):
        self.go_sales = fetch_go_sales_dataset()
        self.go_sales_spark = fetch_go_sales_dataset("spark")

    def test_split_pandas(self):
        train, test, train_y, test_y = multitable_train_test_split(
            self.go_sales,
            main_table_name="go_products",
            label_column_name="Product number",
            test_size=0.2,
        )
        main_table_df: pd.Dataframe = None
        for df in train:
            if get_table_name(df) == "go_products":
                main_table_df = df
        self.assertEqual(len(main_table_df), 220)
        self.assertEqual(len(train_y), 220)
        for df in test:
            if get_table_name(df) == "go_products":
                main_table_df = df
        self.assertEqual(len(main_table_df), 54)
        self.assertEqual(len(test_y), 54)

    def test_split_pandas_1(self):
        train, test, train_y, test_y = multitable_train_test_split(
            self.go_sales,
            main_table_name="go_products",
            label_column_name="Product number",
            test_size=200,
        )
        main_table_df: pd.Dataframe = None
        for df in test:
            if get_table_name(df) == "go_products":
                main_table_df = df
        self.assertEqual(len(main_table_df), 200)
        self.assertEqual(len(test_y), 200)

    def test_split_spark(self):
        train, test, train_y, test_y = multitable_train_test_split(
            self.go_sales_spark,
            main_table_name="go_products",
            label_column_name="Product number",
            test_size=0.2,
        )
        main_table_df: pd.Dataframe = None
        for df in train:
            if get_table_name(df) == "go_products":
                main_table_df = df
        self.assertEqual(main_table_df.count(), 220)
        self.assertEqual(train_y.count(), 220)
        for df in test:
            if get_table_name(df) == "go_products":
                main_table_df = df
        self.assertEqual(main_table_df.count(), 54)
        self.assertEqual(test_y.count(), 54)


class TestConvert(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        targets = ["pandas", "spark", "spark-with-index"]
        cls.tgt2datasets = {tgt: fetch_go_sales_dataset(tgt) for tgt in targets}

    def _check(self, src, dst, tgt):
        self.assertEqual(get_table_name(src), get_table_name(dst), tgt)
        self.assertEqual(list(get_columns(src)), list(get_columns(dst)), tgt)
        pd_src = _ensure_pandas(src)
        pd_dst = _ensure_pandas(dst)
        self.assertEqual(pd_src.shape, pd_dst.shape, tgt)

    def test_to_pandas(self):
        for tgt, datasets in self.tgt2datasets.items():
            transformer = Convert(astype="pandas")
            go_products = datasets[3]
            self.assertEqual(get_table_name(go_products), "go_products", tgt)
            transformed_df = transformer.transform(go_products)
            self.assertTrue(_is_pandas_df(transformed_df), tgt)
            self._check(go_products, transformed_df, tgt)

    def test_to_spark(self):
        for tgt, datasets in self.tgt2datasets.items():
            transformer = Convert(astype="spark")
            go_products = datasets[3]
            self.assertEqual(get_table_name(go_products), "go_products", tgt)
            transformed_df = transformer.transform(go_products)
            self.assertTrue(_is_spark_df(transformed_df), tgt)
            self._check(go_products, transformed_df, tgt)

    def test_to_spark_with_index(self):
        for tgt, datasets in self.tgt2datasets.items():
            transformer = Convert(astype="spark-with-index")
            go_products = datasets[3]
            self.assertEqual(get_table_name(go_products), "go_products", tgt)
            transformed_df = transformer.transform(go_products)
            self.assertTrue(_is_spark_with_index(transformed_df), tgt)
            self._check(go_products, transformed_df, tgt)



# Copyright 2021, 2022 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

import lale.docstrings
import lale.operators

from .project import Project


class _SplitXyImpl:
    def __init__(self, label_name="y"):
        self.label_name = label_name
        self._project_X = None
        self._project_y = None

    def _extract_y(self, X):
        if self._project_y is None:
            self._project_y = Project(columns=[self.label_name])
        result = self._project_y.transform(X)
        if isinstance(result, pd.DataFrame):
            result = result.squeeze()
        return result

    def transform(self, X):
        if self._project_X is None:
            self._project_X = Project(drop_columns=[self.label_name])
        return self._project_X.transform(X)

    def transform_X_y(self, X, y):
        return self.transform(X), self._extract_y(X)

    def viz_label(self) -> str:
        return "SplitXy:\n" + self.label_name


_hyperparams_schema = {
    "description": "The SplitXy operator separates the label field/column from the input dataframe X.",
    "allOf": [
        {
            "description": "This first object lists all constructor arguments with their types, but omits constraints for conditional hyperparameters",
            "type": "object",
            "additionalProperties": False,
            "relevantToOptimizer": [],
            "properties": {
                "label_name": {
                    "description": "The name of the label column in the input dataframe X.",
                    "default": "y",
                    "type": "string",
                },
            },
        }
    ],
}

_input_transform_schema = {
    "type": "object",
    "required": ["X"],
    "additionalProperties": True,
    "properties": {
        "X": {
            "description": "Features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"laleType": "Any"}},
        },
    },
}

_output_transform_schema = {
    "description": "Output data schema for transformed data.",
    "type": "array",
    "items": {"type": "array", "items": {"laleType": "Any"}},
}

_input_transform_X_y_schema = {
    "type": "object",
    "required": ["X", "y"],
    "additionalProperties": False,
    "properties": {
        "X": {
            "description": "Input features; the outer array is over samples.",
            "type": "array",
            "items": {"type": "array", "items": {"laleType": "Any"}},
        },
        "y": {
            "description": "Input labels; ignored.",
            "laleType": "Any",
        },
    },
}

_output_transform_X_y_schema = {
    "type": "array",
    "laleType": "tuple",
    "items": [
        {
            "description": "X",
            "type": "array",
            "items": {"type": "array", "items": {"laleType": "Any"}},
        },
        {
            "description": "y",
            "type": "array",
            "items": {"laleType": "Any"},
        },
    ],
}

_combined_schemas = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "description": "Relational algebra SplitXy operator.",
    "documentation_url": "https://lale.readthedocs.io/en/latest/modules/lale.lib.rasl.split_xy.html",
    "type": "object",
    "tags": {"pre": [], "op": ["transformer"], "post": []},
    "properties": {
        "hyperparams": _hyperparams_schema,
        "input_transform": _input_transform_schema,
        "output_transform": _output_transform_schema,
        "input_transform_X_y": _input_transform_X_y_schema,
        "output_transform_X_y": _output_transform_X_y_schema,
    },
}

SplitXy = lale.operators.make_operator(_SplitXyImpl, _combined_schemas)

lale.docstrings.set_docstrings(SplitXy)



# Copyright 2021 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common interface to manipulate different type of dataframes supported in Lale.
"""

from typing import List, Union

import numpy as np
import pandas as pd

from lale.helpers import (
    _is_pandas_df,
    _is_pandas_series,
    _is_spark_df,
    _is_spark_with_index,
)

column_index = Union[str, int]


def get_columns(df) -> List[column_index]:
    if _is_pandas_series(df):
        return pd.Series([df.name])
    if _is_pandas_df(df):
        return df.columns
    if _is_spark_with_index(df):
        return pd.Series(df.columns_without_indexes)
    if _is_spark_df(df):
        return df.columns
    if isinstance(df, np.ndarray):
        # should have more asserts here
        _, num_cols = df.shape
        return list(range(num_cols))
    assert False, type(df)


def select_col(df, col: column_index):
    if isinstance(df, np.ndarray):
        return df[:, col]  # type: ignore
    elif _is_pandas_df(df):
        return df[col]
    elif _is_spark_df(df):
        return df.select(col)
    else:
        raise ValueError(f"Unsupported series type {type(df)}")


def count(df):
    if isinstance(df, np.ndarray):
        return df.size
    if _is_pandas_df(df) or _is_pandas_series(df):
        return len(df)
    elif _is_spark_df(df):
        return df.count()
    else:
        return len(df)


def make_series_distinct(df):
    if isinstance(df, np.ndarray):
        return np.unique(df)
    elif isinstance(df, pd.Series):
        return df.unique()
    elif _is_spark_df(df):
        return df.distinct()
    else:
        raise ValueError(f"Unsupported series type {type(df)}")


def make_series_concat(df1, df2):
    if isinstance(df1, np.ndarray):
        assert isinstance(df2, np.ndarray)
        return np.concatenate((df1, df2))
    elif isinstance(df1, pd.Series):
        assert isinstance(df2, pd.Series)
        return pd.concat([df1, df2])
    elif _is_spark_df(df1):
        assert _is_spark_df(df2)
        return df1.union(df2)
    else:
        raise ValueError(f"Unsupported series type {type(df1)}")
